# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import dash
from dash import dcc, html
from dash.dependencies import Input, Output, State
import plotly.io as pio

import odatix.gui.themes as themes
import odatix.components.motd as motd

from odatix.gui.css_helper import Style

top_bar_height = "50px"
side_bar_width = "0px"

topbar_pages = ["Architectures", "Workspace"]

def top_bar(gui):
    version = motd.read_version()

    return html.Div(
        children=[
            dcc.Link(
                id="navbar-title",
                className="link",
                href="/",
                children=[
                    html.Span(
                        html.Div(
                            children=[
                                html.Span("Odatix " + str(version), className="link-title1 title"),
                                html.Span("Home", className="link-title2 title"),
                            ],
                            className="link-container"
                        ),
                        className="mask"
                    )
                ],
                style={"position": "block", "marginLeft": "30px", "left": "75px", "zIndex": "2", "transition": "marginLeft 0.25s"},
            ),
            html.Div([
                html.Div(
                    [
                        html.Div(
                            dcc.Link(f"{page['name']}", href=page["relative_path"], className="nav-link"),
                        ) for page in dash.page_registry.values() if page["name"] in topbar_pages
                    ],
                    className="nav-links",
                ),
                html.Div(
                    children=[
                        dcc.Dropdown(
                            id="theme-dropdown",
                            options=[{"label": f"{theme}", "value": f"{theme}"} for theme in themes.list],
                            value=gui.start_theme,
                            className="theme-dropdown",
                            clearable=False,
                            style={"width": "150px", "marginRight": "20px", "marginTop": "3px"},
                        )
                    ],
                    className="tooltip delay bottom auto",
                    **{"data-tooltip": "Select Theme"},
                ),
            ],
            id="nav-right",
            style={"display": "flex", "position": "absolute", "right": "0", "alignItems": "center", "justifyContent": "right", "zIndex": "1000"},)
        ],
        style={"height": f"{top_bar_height}"},
        className="navbar",
        id="navbar",
    )


def side_bar(gui):

    return html.Div(
        [
            html.Div(
                id="sidebar-top",
                className="sidebar-top",
                style={"top": "0px", "left": "0px", "width": side_bar_width, "height": f"{top_bar_height}"},
            ),
            html.Img(
                id="toggle-button",
                className="sidebar-button",
                src="/assets/icons/sidebar-panel-expand-icon.svg",
                n_clicks=0,
                style={
                    "display": "none",
                    "cursor": "pointer",
                    "position": "absolute",
                    "top": "10px",
                    "left": "20px",
                    "width": "30px",
                    "zIndex": "3",
                },
            ),
        ]
    )


def setup_callbacks(gui):
    @gui.app.callback(
        Output("theme", "className"),
        Input("theme-dropdown", "value"),
    )
    def update_theme(
        theme
    ):
        return f"theme {theme}"
    
    @gui.app.callback(
        Output("url", "href"),
        Input({"type": "update_url", "id": dash.ALL}, "data"),
    )
    def update_url(data_list):
        triggered_id = dash.callback_context.triggered_id
        if triggered_id and isinstance(triggered_id, dict):
            if triggered_id.get("type") == "update_url":
                url_trigger_id = triggered_id.get("id")
                for data in data_list:
                    if data and isinstance(data, dict):
                        if data.get("id") == url_trigger_id:
                            return data.get("href", "")
        return dash.no_update



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import dash
import shutil
import uuid
from dash import html, dcc, Input, Output, ctx, State

import odatix.gui.ui_components as ui
from odatix.gui.icons import icon
import odatix.gui.navigation as navigation
import odatix.components.workspace as workspace
import odatix.lib.hard_settings as hard_settings
from odatix.lib.settings import OdatixSettings

page_path = "/architectures"

dash.register_page(
    __name__,
    path=page_path,
    title='Odatix - Architectures',
    name='Architectures',
    order=2,
)


######################################
# UI Components
######################################

def normal_card(name, card_type: str = "arch"):
    unique_key = str(uuid.uuid4())
    return html.Div(
        [
            html.Div(name, style={"fontWeight": "bold", "fontSize": "1.2em", "textAlign": "center"}),
            html.Div([
                html.Div([
                    ui.icon_button(
                        id=f"button-edit-{card_type}-{name}",
                        icon=icon("gear", className="icon black"),
                        text="Settings",
                        color="default",
                        link=f"/{card_type}_editor?{card_type}={name}",
                        width="100px",
                    ),
                    ui.icon_button(
                        id=f"button-open-{card_type}-{name}",
                        icon=icon("edit", className="icon black"),
                        text="Edit Configs",
                        color="default",
                        link=f"/config_editor?{card_type}={name}",
                        multiline=True,
                        width="100px",
                    ),
                ], style={"display": "flex"}),
                html.Div([
                    ui.duplicate_button(id={"type": "button-duplicate", "card_type": card_type, "name": name}),
                    ui.delete_button(id={"type": "button-delete", "card_type": card_type, "name": name}),
                ], style={"display": "flex", "alignItems": "center"}),
            ], style={
                "marginTop": "8px",
                "display": "flex",
                "flexDirection": "row",
                "width": "100%",
                "justifyContent": "space-between",
            }),
        ],
        className="card",
        style={
            "padding": "18px",
            "margin": "10px",
            "width": "300px",
            "height": "100px",
        },
        key=unique_key
    )

def add_card(text: str, card_type: str = "arch"):
    btn_id = {"type": "button-add", "card_type": card_type}
    return html.Div(
        html.Div(
            html.Div(
                children=[
                    html.Div(text, style={"fontWeight": "bold", "fontSize": "1.2em", "color": "var(--add-card-text-color)"}),
                    html.Div(
                        "+",
                        style={
                            "fontSize": "2.5em",
                            "lineHeight": "80px",
                            "height": "80px",
                        }
                    ),
                ], 
                style={"textAlign": "center"}
            ),
            id=btn_id,
            n_clicks=0,
            style={"textDecoration": "none", "color": "var(--add-card-text-color)"},
        ),
        className="card add hover",
        style={
            "padding": "18px",
            "margin": "10px",
            "width": "300px",
            "height": "100px",
        },
    )


######################################
# Callbacks
######################################

# Update cards on page load
@dash.callback(
    Output("arch-cards-matrix", "children"),
    Output("sim-cards-matrix", "children"),
    Input("arch-cards-matrix", "children"),
    State("odatix-settings", "data"),
)
def update_cards(_, odatix_settings):
    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    sim_path = odatix_settings.get("sim_path", OdatixSettings.DEFAULT_SIM_PATH)

    architectures = workspace.get_architectures(arch_path)
    simulations = workspace.get_simulations(sim_path)

    arch_cards = [normal_card(name, "arch") for name in architectures]
    arch_cards.append(add_card("Create New Architecture", "arch"))
    # sim_cards = [normal_card(name, "sim") for name in simulations]
    # sim_cards.append(add_card("Create New Simulation", "sim"))
    sim_cards = []
    return arch_cards, sim_cards

@dash.callback(
    Output("arch-cards-matrix", "children", allow_duplicate=True),
    Output("sim-cards-matrix", "children", allow_duplicate=True),
    Input({"type": "button-duplicate", "card_type": dash.ALL, "name": dash.ALL}, "n_clicks_timestamp"),
    State({"type": "button-duplicate", "card_type": dash.ALL, "name": dash.ALL}, "id"),
    State("odatix-settings", "data"),
    prevent_initial_call=True
)
def direct_duplicate(dupl_timestamps, btn_ids, odatix_settings):
    triggered = ctx.triggered_id
    if not triggered or not isinstance(triggered, dict):
        return dash.no_update, dash.no_update

    if not dupl_timestamps or not btn_ids:
        return dash.no_update, dash.no_update
    
    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    sim_path = odatix_settings.get("sim_path", OdatixSettings.DEFAULT_SIM_PATH)

    idx = max(range(len(dupl_timestamps)), key=lambda i: dupl_timestamps[i] or 0)
    btn_id = btn_ids[idx]

    if btn_id != triggered or not dupl_timestamps[idx]:
        return dash.no_update, dash.no_update

    card_type = btn_id["card_type"]
    name = btn_id["name"]

    path = arch_path if card_type == "arch" else sim_path

    base = name
    suffix = 1
    while True:
        new_name = f"{base}_copy{suffix}"
        if not workspace.instance_exists(path, new_name):
            break
        suffix += 1
        if suffix > 1000:
            return dash.no_update, dash.no_update

    try:
        workspace.duplicate_instance(path, name, new_name)
    except Exception:
        return dash.no_update, dash.no_update

    # Refresh the cards
    architectures = workspace.get_architectures(arch_path)
    simulations = workspace.get_simulations(sim_path)
    arch_cards = [normal_card(n, "arch") for n in architectures]
    arch_cards.append(add_card("Create New Architecture", "arch"))
    sim_cards = [normal_card(n, "sim") for n in simulations]
    sim_cards.append(add_card("Create New Simulation", "sim"))
    return arch_cards, sim_cards

# Open deletion popup
@dash.callback(
    Output("delete-popup", "className"),
    Output("delete-popup-message", "children"),
    Output("delete-info", "data"),
    Input({"type": "button-delete", "card_type": dash.ALL, "name": dash.ALL}, "n_clicks"),
    State({"type": "button-delete", "card_type": dash.ALL, "name": dash.ALL}, "id"),
    prevent_initial_call=True
)
def show_delete_popup(n_clicks, btn_ids):
    if not ctx.triggered_id or not isinstance(ctx.triggered_id, dict) or all(n == 0 for n in n_clicks):
        return dash.no_update, dash.no_update, dash.no_update
    
    card_type = ctx.triggered_id["card_type"]
    name = ctx.triggered_id["name"]
    msg = f"Do you really want to delete {card_type} '{name}'?"
    return "overlay-odatix visible", msg, {"card_type": card_type, "name": name}

# Popup close
@dash.callback(
    Output("delete-popup", "className", allow_duplicate=True),
    Input("delete-cancel-btn", "n_clicks"),
    prevent_initial_call=True
)
def close_delete_popup(n):
    return "overlay-odatix"

# Delete
@dash.callback(
    Output("delete-popup", "className", allow_duplicate=True),
    Output("delete-error", "children"),
    Output("arch-cards-matrix", "children", allow_duplicate=True),
    Output("sim-cards-matrix", "children", allow_duplicate=True),
    Input("delete-confirm-btn", "n_clicks"),
    State("delete-info", "data"),
    State("odatix-settings", "data"),
    prevent_initial_call=True
)
def do_delete(n_clicks, info, odatix_settings):
    if not n_clicks or not info:
        return dash.no_update, dash.no_update, dash.no_update, dash.no_update
    
    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    sim_path = odatix_settings.get("sim_path", OdatixSettings.DEFAULT_SIM_PATH)
    
    card_type = info["card_type"]
    name = info["name"]
    path = arch_path if card_type == "arch" else sim_path
    if not workspace.instance_exists(path, name):
        return dash.no_update, "Item not found.", dash.no_update, dash.no_update
    try:
        workspace.delete_instance(path, name)
    except Exception as e:
        print("Error during deletion:", e)
        return dash.no_update, f"Error: {e}", dash.no_update, dash.no_update

    # Refresh the cards
    architectures = workspace.get_architectures(arch_path)
    simulations = workspace.get_simulations(sim_path)
    arch_cards = [normal_card(name, "arch") for name in architectures]
    arch_cards.append(add_card("Create New Architecture", "arch"))
    sim_cards = [normal_card(name, "sim") for name in simulations]
    sim_cards.append(add_card("Create New Simulation", "sim"))
    return "overlay-odatix", "", arch_cards, sim_cards

@dash.callback(
    Output({"type": "update_url", "id": page_path}, "data"),
    Input({"type": "button-add", "card_type": dash.ALL}, "n_clicks_timestamp"),
    State({"type": "button-add", "card_type": dash.ALL}, "id"),
    State("odatix-settings", "data"),
    prevent_initial_call=True
)
def handle_add_card(n_clicks_timestamps, ids, odatix_settings):
    if not n_clicks_timestamps or all(ts is None for ts in n_clicks_timestamps):
        return dash.no_update

    triggered_id = ctx.triggered_id
    if not triggered_id or not isinstance(triggered_id, dict):
        return dash.no_update
    
    card_type = triggered_id.get("card_type")

    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    sim_path = odatix_settings.get("sim_path", OdatixSettings.DEFAULT_SIM_PATH)
    root = arch_path if card_type == "arch" else sim_path

    base_name = "New_Architecture" if card_type == "arch" else "New_Simulation"

    # Search for an available name
    for i in range(1, 1001):
        candidate = f"{base_name}{i}"
        if not workspace.instance_exists(root, candidate):
            return {"href": f"/{card_type}_editor?{card_type}={candidate}", "id":page_path}
    # If all names are taken, go to editor without preset name
    return {"href": f"/{card_type}_editor", "id":page_path}


######################################
# Layout
######################################

layout = html.Div(
    children=[
        dcc.Location(id="url", refresh='callback-nav'),
        html.Div(
            children=[
                html.H2("Architectures", style={"textAlign": "center"}),
                html.Div(id="arch-cards-matrix", className="card-matrix"),
                # html.H2("Simulations", style={"textAlign": "center", "marginTop": "40px"}),
                html.Div(id="sim-cards-matrix", className="card-matrix"),
            ],
            style={
                "display": "block",
                "width": "auto",
                "textAlign": "center",
                "marginBottom": "10px",
            },
        ),
        dcc.Store(id="duplicate-info"),
        html.Div(
            id="duplicate-popup",
            className="overlay-odatix",
            children=[
                html.Div([
                    html.H3("Duplicate"),
                    html.Button("×", id="duplicate-cancel-btn", n_clicks=0, className="close"),
                    html.Div(id="duplicate-popup-message"),
                    dcc.Input(id="duplicate-new-name", placeholder="New name", type="text", style={"width": "90%", "marginTop": "10px"}),
                    html.Div([
                        html.Button("Create", id="duplicate-create-btn", n_clicks=0, style={"marginRight": "10px"}),
                    ], style={"marginTop": "18px"}),
                    html.Div(id="duplicate-error", style={"color": "red", "marginTop": "10px"}),
                ], className="popup-odatix")
            ]
        ),
        dcc.Store(id="delete-info"),
        html.Div(
            id="delete-popup",
            className="overlay-odatix",
            children=[
                html.Div([
                    html.H2("Warning"),
                    html.Div(id="delete-popup-message"),
                    html.Div("This action is irreversible.", style={"marginTop": "10px", "color": "#FA5252", "fontWeight": "bold"}),
                    html.Div([
                        ui.icon_button(
                            icon=icon("delete", className="icon"),
                            color="caution", 
                            text="Delete", 
                            width="90px",
                            id="delete-confirm-btn",
                        ),
                        html.Button("Cancel", id="delete-cancel-btn", n_clicks=0, style={"marginLeft": "10px", "width": "90px"}),
                    ], style={"marginTop": "18px", "display": "flex", "justifyContent": "center"}),
                    html.Div(id="delete-error", style={"color": "red", "marginTop": "10px"}),
                ], className="popup-odatix")
            ]
        ),
        dcc.Store(id={"type": "update_url", "id": page_path}),
    ],
    className="page-content",
    style={
        "padding": "0 16%",
        "display": "flex",  
        "flexDirection": "column",
        "min-height": f"calc(100vh - {navigation.top_bar_height})",
    },
)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import dash
from dash import dcc, html
from dash.dependencies import Input, Output
import plotly.graph_objs as go

import odatix.explorer.legend as legend
import odatix.explorer.content_lib as content_lib
import odatix.explorer.figures as figures
import odatix.explorer.themes as themes

def setup_callbacks(explorer, all_checklist_inputs, all_architecture_inputs, all_target_inputs, all_domains_inputs):

  @explorer.app.callback(
    Output("graph-scatter", "children"),
    [
      Input("yaml-dropdown", "value"),
      Input("metric-x-dropdown", "value"),
      Input("metric-y-dropdown", "value"),
      Input("results-dropdown", "value"),
      Input("show-all-architectures", "n_clicks"),
      Input("hide-all-architectures", "n_clicks"),
      Input("toggle-legend", "value"),
      Input("toggle-legendgroup", "value"),
      Input("toggle-title", "value"),
      Input("toggle-lines-scatter", "value"),
      Input("toggle-labels", "value"),
      Input("color-mode-dropdown", "value"),
      Input("symbol-mode-dropdown", "value"),
      Input("dl-format-dropdown", "value"),
      Input("background-dropdown", "value"),
      Input("theme-dropdown", "value"),
      Input("toggle-unique-architectures", "value"),
      Input("toggle-unique-targets", "value"),
      Input("dissociate-domain-dropdown", "value"),
    ]
    + all_checklist_inputs  
  )
  def update_graph(
    selected_yaml,
    selected_metric_x,
    selected_metric_y,
    selected_results,
    show_all,
    hide_all,
    toggle_legend,
    toggle_legendgroup,
    toggle_title,
    toggle_lines,
    toggle_labels,
    color_mode,
    symbol_mode,
    dl_format,
    background,
    theme,
    toggle_unique_architectures,
    toggle_unique_targets,
    dissociate_domain,
    *checklist_values,
  ):
    try:
      arch_checklist_values = checklist_values[:len(all_architecture_inputs)]
      target_checklist_values = checklist_values[len(all_architecture_inputs):len(all_architecture_inputs) + len(all_target_inputs)]
      domain_checklist_values = checklist_values[len(all_architecture_inputs) + len(all_target_inputs):]

      if not selected_yaml or selected_yaml not in explorer.dfs:
        return html.Div(className="error", children=[html.Div("Please select a YAML file.")])

      selected_metric_x_display = selected_metric_x.replace("_", " ") if selected_metric_x is not None else ""
      selected_metric_y_display = selected_metric_y.replace("_", " ") if selected_metric_y is not None else ""

      unit_x = legend.unit_to_html(explorer.units[selected_yaml].get(selected_metric_x, ""))
      unit_y = legend.unit_to_html(explorer.units[selected_yaml].get(selected_metric_y, ""))

      selected_metric_x_display_unit = (
        selected_metric_x_display + " (" + unit_x + ")" if unit_x else selected_metric_x_display
      )
      selected_metric_y_display_unit = (
        selected_metric_y_display + " (" + unit_y + ")" if unit_y else selected_metric_y_display
      )

      ctx = dash.callback_context
      triggered_id = ctx.triggered[0]["prop_id"].split(".")[0]

      if triggered_id in ["show-all-architectures", "hide-all-architectures"]:
        visible_architectures = set(
          explorer.dfs[selected_yaml]["Architecture"].unique() if triggered_id == "show-all-architectures" else []
        )
      else:
        visible_architectures = set(
          architecture for i, architecture in enumerate(explorer.all_architectures) if arch_checklist_values[i] and architecture in explorer.dfs[selected_yaml]["Architecture"].unique()
        )

      if triggered_id in ["show-all-targets", "hide-all-targets"]:
        visible_targets = set(
          explorer.dfs[selected_yaml]["Target"].unique() if triggered_id == "show-all-targets" else []
        )
      else:
        visible_targets = set(
          target for i, target in enumerate(explorer.all_targets) if target_checklist_values[i] and target in explorer.dfs[selected_yaml]["Target"].unique()
        )
      
      i_current_value = 0
      domains = {}
      for domain in explorer.all_param_domains.keys():
        domains[domain] = {}
        for config in explorer.all_param_domains[domain]:
          visible = True if domain_checklist_values[i_current_value] else False
          domains[domain].update({f'{config}': visible})
          i_current_value += 1

      filtered_df = explorer.dfs[selected_yaml][
        explorer.dfs[selected_yaml]["Architecture"].isin(visible_architectures)
      ]

      for domain, configs in domains.items():
        if domain in filtered_df.columns:
          filtered_df.loc[:, domain] = filtered_df[domain].fillna("None")
          excluded_values = [config for config, is_visible in configs.items() if not is_visible]

          # Apply exclusion only if there are forbidden values in the column
          if any(filtered_df[domain].isin(excluded_values)):
            filtered_df = filtered_df[~filtered_df[domain].isin(excluded_values)]

      fig = go.Figure()

      i_target = -1
      for i_unique_target, target in enumerate(explorer.all_targets):
        if target in explorer.dfs[selected_yaml]["Target"].unique():
          i_target = i_target + 1
        if target not in visible_targets:
          continue

        i_architecture = -1
        for i_unique_architecture, architecture in enumerate(explorer.all_architectures):
          if architecture in explorer.dfs[selected_yaml]["Architecture"].unique():
            i_architecture = i_architecture + 1
          if architecture in visible_architectures:
            df_architecture_target = filtered_df[
              (filtered_df["Architecture"] == architecture) & 
              (filtered_df["Target"] == target)
            ]
            if df_architecture_target.empty:
              continue

            mode = "lines+markers" if toggle_lines else "markers"

            if dissociate_domain != "None" and dissociate_domain in df_architecture_target.columns:
              dissociation_values = df_architecture_target[dissociate_domain].dropna().unique()
              # cleaned_configurations = [legend.clean_configuration_name(cfg, dissociate_domain) for cfg in unique_configurations]
            else:
              dissociation_values = [None]
              # cleaned_configurations = unique_configurations

            i_dissociate_domain = -1
            for dissociation_value in dissociation_values:
              df_filtered = df_architecture_target.copy()
              if dissociation_value is not None and dissociation_value != "None":
                df_filtered = df_filtered[df_filtered[dissociate_domain] == dissociation_value]
                architecture_diplay = architecture + f" [{dissociate_domain.replace('__main__', 'main')}:{dissociation_value}]" 
              else:
                architecture_diplay = architecture
              i_dissociate_domain += 1

              df_fmax = df_filtered[df_filtered["Type"] == "Fmax"]
              if selected_results in ["All", "Fmax"] and not df_fmax.empty:
                if selected_metric_x is None or selected_metric_x not in df_fmax.columns:
                  return html.Div(className="error", children=[html.Div("Please select a valid x metric.")])
                if selected_metric_y is None or selected_metric_y not in df_fmax.columns:
                  return html.Div(className="error", children=[html.Div("Please select a valid y metric.")])

                x_values = df_fmax[selected_metric_x].tolist()
                y_values = df_fmax[selected_metric_y].tolist()
                config_names = df_fmax["Configuration"].tolist()
                cleaned_config_names = [legend.clean_configuration_name(cfg, dissociate_domain) for cfg in config_names]
                targets = [target] * len(x_values)
                frequencies = ["fmax"] * len(x_values)

                if color_mode == "architecture":
                  color_id = i_unique_architecture if toggle_unique_architectures else i_architecture
                elif color_mode == "target":
                  color_id = i_unique_target if toggle_unique_targets else i_target
                elif color_mode == "domain_value":
                  color_id = i_dissociate_domain if dissociation_value != "None" else -1
                else:
                  color_id = 0

                if symbol_mode == "none":
                  symbol_id = 0
                elif symbol_mode == "architecture":
                  symbol_id = i_unique_architecture if toggle_unique_architectures else i_architecture
                elif symbol_mode == "target":
                  symbol_id = i_unique_target if toggle_unique_targets else i_target
                elif symbol_mode == "domain_value":
                  symbol_id = i_dissociate_domain
                else:
                  symbol_id = 0

                if toggle_labels:
                  mode += "+text"

                figures.add_trace_to_scatter_fig(
                  fig, x_values, y_values, mode, architecture_diplay, frequencies, "fmax", cleaned_config_names,
                  targets, target, selected_metric_x_display, selected_metric_y_display,
                  unit_x, unit_y, color_id, symbol_id, toggle_lines, toggle_legendgroup
                )

              df_range = df_filtered[df_filtered["Type"] == "Custom Freq"]
              if selected_results in ["All", "Custom Freq"] and not df_range.empty:
                for i_freq, frequency in enumerate(explorer.all_frequencies):
                  df_frequency = df_range[df_range["Frequency"] == frequency]
                  if df_frequency.empty:
                    continue

                  if selected_metric_x is None or selected_metric_x not in df_frequency.columns:
                    return html.Div(className="error", children=[html.Div("Please select a valid x metric.")])
                  if selected_metric_y is None or selected_metric_y not in df_frequency.columns:
                    return html.Div(className="error", children=[html.Div("Please select a valid y metric.")])

                  x_values = df_frequency[selected_metric_x].tolist()
                  y_values = df_frequency[selected_metric_y].tolist()
                  config_names = df_frequency["Configuration"].tolist()
                  cleaned_config_names = [legend.clean_configuration_name(cfg, dissociate_domain) for cfg in config_names]
                  targets = [target] * len(x_values)
                  frequencies = [f"{frequency} MHz"] * len(x_values)

                  if color_mode == "architecture":
                    color_id = i_architecture
                  elif color_mode == "target":
                    color_id = i_target
                  elif color_mode == "domain_value":
                    color_id = i_dissociate_domain if dissociation_value != "None" else -1
                  else:
                    color_id = i_freq + 1
                    
                  if symbol_mode == "none":
                    symbol_id = 0
                  elif symbol_mode == "architecture":
                    symbol_id = i_unique_architecture if toggle_unique_architectures else i_architecture
                  elif symbol_mode == "target":
                    symbol_id = i_unique_target if toggle_unique_targets else i_target
                  elif symbol_mode == "domain_value":
                    symbol_id = i_dissociate_domain
                  else:
                    symbol_id = i_freq + 1

                  if toggle_labels:
                    mode += "+text"

                  figures.add_trace_to_scatter_fig(
                    fig, x_values, y_values, mode, architecture_diplay, frequencies, f"{frequency} MHz", cleaned_config_names,
                    targets, target, selected_metric_x_display, selected_metric_y_display,
                    unit_x, unit_y, color_id, symbol_id, toggle_lines, toggle_legendgroup
                  )

      fig.update_layout(
        paper_bgcolor=background,
        showlegend=True if toggle_legend else False,
        legend_groupclick="toggleitem",
        xaxis_title=selected_metric_x_display_unit,
        yaxis_title=selected_metric_y_display_unit,
        xaxis=dict(range=[0, None]),
        yaxis=dict(range=[0, None]),
        title=selected_metric_y_display + " vs " + selected_metric_x_display if toggle_title else None,
        title_x=0.5,
        autosize=True,
        template=theme,
        modebar={
          "bgcolor": themes.get_page_bgcolor(theme, default=None),
          "color": themes.get_button_color(theme, default=None),
          "activecolor": themes.get_button_active_color(theme, default=None),
        }
      )
      filename = "Odatix-{}-{}-vs-{}".format(
        os.path.splitext(selected_yaml)[0], selected_metric_y, selected_metric_x
      )
      return html.Div(
        [
          dcc.Graph(
            figure=fig,
            style={"width": "100%", "height": "100%"},
            config={
              "displayModeBar": True,
              "displaylogo": False,
              "modeBarButtonsToRemove": ["lasso", "select"],
              "toImageButtonOptions": {
                "format": dl_format,
                "scale": 3,
                "filename": filename,
              },
            },
          )
        ],
        style={"width": "100%", "height": "100%", "display": "inline-block", "vertical-align": "top"},
      )
    except Exception as e:
      return content_lib.generate_error_div(e)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import dash
from dash import html, dcc, Input, Output, State, ctx
import uuid
import random

import odatix.gui.ui_components as ui
from odatix.gui.utils import get_key_from_url
import odatix.gui.navigation as navigation
import odatix.lib.hard_settings as hard_settings
from odatix.lib.settings import OdatixSettings
import odatix.components.replace_params as replace_params
import odatix.components.workspace as workspace
from odatix.gui.icons import icon
from odatix.gui.css_helper import Style

verbose = False

page_path = "/config_editor"

dash.register_page(
    __name__,
    path=page_path,
    title='Odatix - Configuration Editor',
    name='Configuration Editor',
    order=4,
)

######################################
# Helper Functions
######################################

def split_by_domain(flat_list, lengths):
    result = []
    idx = 0
    for l in lengths:
        result.append(flat_list[idx:idx+l])
        idx += l
    return result

def get_uuid():
    return str(uuid.uuid4())

def generate_config_link(arch_name: str = "", domain_name: str = "") -> str:
    if domain_name:
        return f"/config_generator?arch={arch_name}&domain={domain_name}"
    else:
        return f"/config_generator?arch={arch_name}"

######################################
# UI Components
######################################

def config_card(domain_uuid, config_uuid, config_name, content, initial_content, config_layout="normal"):
    display_name = config_name[:-4] if config_name.endswith(".txt") else config_name

    save_class =  "color-button disabled"
    status_text = ""
    status_class = "status"
    return html.Div([
        html.Div([
            dcc.Input(
                value=f"{display_name}",
                type="text",
                id={"type": "config-title", "domain_uuid": domain_uuid, "config_uuid": config_uuid},
                className="title-input",
                style={
                    "width": "calc(100% - 20px)",
                    "marginLeft": "5px",
                    "marginRight": "5px",
                    "fontWeight": "bold",
                    "fontSize": "1.1em",
                    "height": "10px",
                    "marginTop": "-5px",
                    "marginBottom": "2px",
                    "textAlign": "center",
                },
            )
        ]),
        dcc.Textarea(
            id={"type": "config-content", "domain_uuid": domain_uuid, "config_uuid": config_uuid},
            value=content,
            className="auto-resize-textarea" if config_layout != "compact" else "",
            style={
                "width": "calc(100% - 20px)",
                "marginLeft": "5px",
                "marginRight": "5px",
                "resize": "none" if config_layout != "compact" else "vertical",
                "minHeight": "none" if config_layout != "compact" else "45px",
                "height": "none" if config_layout != "compact" else "45px",
                "fieldSizing": "border-box",
                "fontFamily": "monospace",
                "fontWeight": "normal",
            },
        ),
        dcc.Store(
            id={"type": "config-metadata", "domain_uuid": domain_uuid, "config_uuid": config_uuid},
            data={
                "domain_uuid": domain_uuid,
                "config_uuid": config_uuid,
                "config_name": config_name,
                "config_content": content,
            }
        ),
        html.Div([
            html.Div([
                html.Div([ui.icon_button(
                    icon=icon("save", className="icon", id={"type": "save-config-icon", "domain_uuid": domain_uuid, "config_uuid": config_uuid}),
                    color="disabled",
                    text="Save", 
                    width="78px",
                    id={"type": "save-config", "domain_uuid": domain_uuid, "config_uuid": config_uuid},
                ),], style={"marginLeft": "5px"}),
                html.Div(status_text, id={"type": "save-status", "domain_uuid": domain_uuid, "config_uuid": config_uuid}, className=status_class, style={"marginLeft": "0px", "textWrap": "wrap", "width": "70px", "fontWeight": "515"}),
            ], style={"display": "flex", "alignItems": "center"}),
            html.Div([
                ui.duplicate_button(id={"type": "duplicate-config", "domain_uuid": domain_uuid, "config_uuid": config_uuid}),
                ui.delete_button(id={"type": "delete-config", "domain_uuid": domain_uuid, "config_uuid": config_uuid}),
            ], style={"display": "flex", "alignItems": "center", "marginLeft": "0px"}, className="inline-flex-buttons"),
        ], style={
            "marginTop": "8px",
            "display": "flex",
            "flexDirection": "row",
            "width": "100%",
            "justifyContent": "space-between",
        }),
        dcc.Store(id={"type": "initial-title", "domain_uuid": domain_uuid, "config_uuid": config_uuid}, data=display_name),
        dcc.Store(id={"type": "initial-content", "domain_uuid": domain_uuid, "config_uuid": config_uuid}, data=initial_content),
    ], 
    className="card configs", 
    id={"type": "config-card", "domain_uuid": domain_uuid, "config_uuid": config_uuid},
    style={
        "padding": "10px", 
        "margin": "5px", 
        "display": "inline-block", 
        "verticalAlign": "top"
    })

def add_card(text: str = "Add design configuration", domain_uuid: str = hard_settings.main_parameter_domain):
    return html.Div(
        html.Div(
            html.Div(
                children=[
                    html.Div(text, style={"fontWeight": "bold", "fontSize": "1.2em", "paddingTop": "20px"}),
                    html.Div(
                        "+",
                        style={
                            "fontSize": "2.5em",
                            "lineHeight": "80px",
                            "height": "80px",
                        }
                    ),
                ], 
                style={"textAlign": "center"}
            ),
            id={"type": "new-config", "domain_uuid": domain_uuid},
            n_clicks=0,
            style={"display": "flex", "flexDirection": "column", "alignItems": "center", "justifyContent": "center", "height": "100%"}
        ),
        className=f"card configs add hover",
        id={"type": "add-config-card", "domain_uuid": domain_uuid},
        style={
            "padding": "10px",
            "margin": "5px",
            "display": "inline-block",
            "verticalAlign": "top",
            "boxSizing": "border-box"
        },
    )

def architecture_title(arch_name:str=""):
    title_content = html.Div(
        children=[
            html.H3(arch_name, id=f"main_title", style={"marginBottom": "0px"}),
            html.Div(
                children=[
                    ui.icon_button(
                        id=f"button-open-config-editor",
                        icon=icon("gear", className="icon"),
                        text="Architecture Settings",
                        color="default",
                        link=f"/arch_editor?arch={arch_name}",
                        multiline=True,
                        width="135px",
                    ),
                    dcc.Dropdown(
                        id="config-layout-dropdown", 
                        options=[
                            {"label": "Compact Layout", "value": "compact"},
                            {"label": "Normal Layout", "value": "normal"},
                            {"label": "Wide Layout", "value": "wide"},
                        ],
                        value="normal",
                        clearable=False,
                        style={"width": "155px"},
                    ),
                ],
                className="inline-flex-buttons",
            )
        ],
        className="title-tile-flex",
        style={
            "display": "flex",
            "alignItems": "center",
            "padding": "0px",
            "justifyContent": "space-between",
        }
    )
    back_btn = ui.back_button(link="/architectures")
    return html.Div([
        html.Div(style={"display": "none"}),
        html.Div(style={"display": "none"}),
        html.Div(style={"display": "none"}),
        html.Div(
            children=[
                back_btn,
                title_content
            ],
            className="tile title",
            style={"position": "relative"},
        )],
        className="card-matrix config",
        style={"marginLeft": "-13px", "marginBottom": "0px"},
    )

def parameter_domain_title(domain_name:str=hard_settings.main_parameter_domain, domain_uuid:str=hard_settings.main_parameter_domain, arch_name:str=""):
    if domain_uuid == hard_settings.main_parameter_domain:
        text = "Main parameter domain"
        buttons = html.Div(
            children=[
                ui.icon_button(
                    # id={"type": "generate-config", "domain_uuid": domain_uuid},
                    icon=icon("generate", className="icon"),
                    text="Config Generator",
                    color="default",
                    link=generate_config_link(arch_name=arch_name),
                    multiline=True,
                    tooltip="Generate multiple design configurations",
                ),
                ui.duplicate_button(
                    id={"action": "duplicate-domain", "domain_uuid": domain_uuid},
                    tooltip="Duplicate as a new domain",
                ),
                # ui.icon_button(
                #     id={"action": "duplicate-domain", "domain_uuid": domain_uuid},
                #     icon=icon("duplicate", className="icon"),
                #     text="Duplicate as Domain",
                #     color="secondary",
                #     multiline=True,
                #     width="140px",
                #     tooltip="Duplicate the main parameter domain as a new domain",
                # ),
            ],
            className="inline-flex-buttons",
        )
        tooltip = "The main parameter domain is the default domain used for parameter replacement. Several parameter domains can be created to manage different sets of parameters for different design configurations or scenarios. It mainly serves as a convenient way to organize design configurations and to easily run combinations of design configurations across multiple domains."

        return ui.title_tile(text=text, id=f"domain_title_{domain_uuid}", buttons=buttons, tooltip=tooltip)
    else:
        buttons = html.Div(
            children=[
                ui.icon_button(
                    id={"type": "generate-config", "domain_uuid": domain_uuid},
                    icon=icon("generate", className="icon"),
                    text="Config Generator",
                    color="default",
                    link=generate_config_link(arch_name=arch_name, domain_name=domain_name),
                    multiline=True,
                    tooltip="Generate multiple design configurations",
                ),
                ui.duplicate_button(
                    id={"action": "duplicate-domain", "domain_uuid": domain_uuid},
                    tooltip="Duplicate domain",
                ),
                ui.delete_button(
                    id={"action": "delete-domain", "domain_uuid": domain_uuid},
                    large=False,
                    tooltip="Delete domain",
                )
            ],
            className="inline-flex-buttons param-domain-title",
        )
        title_content = html.Div([
            html.Div([
                html.H3("Parameter domain:", style={"display": "inline-block", "marginBottom": "0px", "marginRight": "10px"}),
                dcc.Input(
                    value=domain_name,
                    type="text",
                    id={"type": "domain-title-input", "domain_uuid": domain_uuid},
                    placeholder="Parameter domain name...",
                    className="title-input domain",
                    style={
                        "marginBottom": "0",
                        "marginTop": "0",
                        "marginRight": "5px",
                        "verticalAlign": "middle",
                        "position": "relative",
                        "top": "-4px",
                    }
                ),
                ui.icon_button(
                    id={"type": "save-domain-title", "domain_uuid": domain_uuid},
                    icon=icon("check", className="icon invisible"),
                    color="invisible",
                    style={"transform": "translateY(6px)"},
                ),
            ],
            style={
                "justifyContent": "flex-start"
            }),
            buttons,
        ],
        className="title-tile-flex",
        style={
            "display": "flex",
            "alignItems": "center",
            "padding": "0px",
            "justifyContent": "space-between",
        }) 
    return html.Div(
        [title_content],
        id={"type": "param-domain-title", "domain_uuid": domain_uuid}, 
        className="tile title",
        style={"marginTop": "50px"} if domain_uuid != hard_settings.main_parameter_domain else {}
    )

def add_parameter_domain_button(text:str="Main parameter domain"):
    return html.Div(
        [
            html.H3(text, style={"marginBottom": "0px"})
        ], 
        id={"type": "button", "action": "add-domain"}, 
        n_clicks=0,
        className="tile title add hover",
        style={
            "marginTop": "50px",
            "textAlign": "center",
            "border": "1px dashed #bbb",
        },
    )

def config_parameters_form(domain_uuid, settings):
    defval = lambda k, v=None: settings.get(k, v)

    save_button = html.Div(
        children=[
            ui.icon_button(
                icon=icon("save", className="icon", id={"type": "save-params-icon", "domain_uuid": domain_uuid}),
                color="disabled",
                text="Save", 
                width="78px",
                id={"type": "save-params-btn", "domain_uuid": domain_uuid},
            ),
        ],
        style={"marginBottom": "10px"},
        className="inline-flex-buttons",
    )
    use_parameters = defval("use_parameters", True)
    return html.Div(
        children=[
            ui.subtitle_div(text="Configuration Parameters", buttons=save_button),
            dcc.Checklist(
                options=[{"label": "Enable parameter replacement", "value": True}],
                value=[True] if use_parameters else [],
                id={"type": "use_parameters", "domain_uuid": domain_uuid},
                className="checklist-switch",
                style={"marginBottom": "12px", "marginTop": "5px"},
            ),
            html.Div(
                children=[
                    html.Div([
                        html.Label("Param Target File"),
                        ui.tooltip_icon("The file used as the target for parameter replacement. This is typically the top-level design file or a design configuration file."),
                        dcc.Input(id={"type": "param_target_file", "domain_uuid": domain_uuid}, value=defval("param_target_file", ""), type="text", placeholder="Top level file used by default", style={"width": "95%"}),
                    ], style={"marginBottom": "12px"}),
                    html.Div([
                        html.Label("Start Delimiter"),
                        ui.tooltip_icon("The delimiter that marks the beginning of a section to be replaced in the target file."),
                        dcc.Input(id={"type": "start_delimiter", "domain_uuid": domain_uuid}, value=defval("start_delimiter", ""), type="text", style={"width": "95%"}),
                    ], style={"marginBottom": "12px"}),
                    html.Div([
                        html.Label("Stop Delimiter"),
                        ui.tooltip_icon("The delimiter that marks the end of a section to be replaced in the target file."),
                        dcc.Input(id={"type": "stop_delimiter", "domain_uuid": domain_uuid}, value=defval("stop_delimiter", ""), type="text", style={"width": "95%"}),
                    ], style={"marginBottom": "12px"}),
                ],
                id={"type": "params-config-fields", "domain_uuid": domain_uuid}, className="animated-section" if use_parameters else "animated-section hide",
            ),
            html.Div(id={"type": "save-params-status", "domain_uuid": domain_uuid}, className="status", style={"marginLeft": "16px"}),
        ]
    )

def preview_pane(domain_uuid:str, settings: dict, domain_settings: dict, replacement_text: str):
    use_parameters = domain_settings.get("use_parameters", True)
    if not use_parameters:
        return preview_div(html.Div("Parameter replacement disabled.", style={"color": "#888"}))
    param_target_file = domain_settings.get("param_target_file", "")
    generate_rtl = settings.get("generate_rtl", False)
    generate_rtl = True if generate_rtl else False # Convert from [True]/[] to True/False
    if generate_rtl:
        base_path = settings.get("design_path", "")
        if not base_path:
            return preview_div(html.Div("No design path specified in architecture settings. Unable to preview.", className="error"))
        if param_target_file == "":
            param_target_file = settings.get("top_level_file", "")
    else:
        rtl_path = settings.get("rtl_path", "")
        if not rtl_path:
            return preview_div(html.Div("No RTL path specified in architecture settings. Unable to preview.", className="error"))
        base_path = rtl_path
        if param_target_file == "":
            param_target_file = os.path.join(settings.get("top_level_file", ""))
    param_target_file = os.path.join(base_path, param_target_file)
    if os.path.isfile(param_target_file):
        base_text = replace_params.read_file(param_target_file)
        start_delimiter = domain_settings.get("start_delimiter", "")
        stop_delimiter = domain_settings.get("stop_delimiter", "")
        new_content, match_found = replace_params.replace_content(
            base_text=base_text,
            replacement_text=replacement_text,
            start_delim=start_delimiter, 
            stop_delim=stop_delimiter, 
            replace_all_occurrences=False,
        )
        if not match_found:
            preview_components=html.Span(base_text, style={"whiteSpace": "pre-wrap", "color": "#FA5252", "fontWeight": "800"}) 
        else:
            start_line, start_charater = replace_params.get_first_appearance(new_content, start_delimiter)
            stop_line, stop_charater = replace_params.get_first_appearance(new_content, stop_delimiter, start_line=start_line, start_char=start_charater+len(start_delimiter)+1)
            lines = new_content.splitlines()
            preview_start = 0
            preview_stop = len(lines)
            preview_lines = lines[preview_start:preview_stop]
            preview_components = []
            # preview_components.append(html.Span("[...]\n", style={"color": "#888"}))
            for idx, line in enumerate(preview_lines):
                line_idx = preview_start + idx
                line_parts = []

                # Case where start and stop are on the same line
                if start_line == stop_line and line_idx == start_line - 1:
                    s_idx = start_charater
                    e_idx = stop_charater
                    if s_idx > 0:
                        line_parts.append(line[:s_idx])
                    line_parts.append(html.Span(start_delimiter, className="text-highlight primary"))
                    line_parts.append(html.Span(line[s_idx+len(start_delimiter):e_idx], className="text-highlight secondary"))
                    line_parts.append(html.Span(stop_delimiter, className="text-highlight primary"))
                    if e_idx + len(stop_delimiter) < len(line):
                        line_parts.append(line[e_idx+len(stop_delimiter):])
                # Start delimiter line
                elif line_idx == start_line - 1:
                    if start_charater != -1:
                        if start_charater > 0:
                            line_parts.append(line[:start_charater])
                        line_parts.append(html.Span(start_delimiter, className="text-highlight primary"))
                        line_parts.append(html.Span(line[start_charater+len(start_delimiter):], className="text-highlight secondary"))
                    else:
                        line_parts.append(line)
                # Stop delimiter line
                elif line_idx == stop_line - 1:
                    if stop_charater != -1:
                        if stop_charater > 0:
                            line_parts.append(html.Span(line[:stop_charater], className="text-highlight secondary"))
                        line_parts.append(html.Span(stop_delimiter, className="text-highlight primary"))
                        if stop_charater + len(stop_delimiter) < len(line):
                            line_parts.append(line[stop_charater+len(stop_delimiter):])
                    else:
                        line_parts.append(line)
                # Replaced content lines
                elif start_line-1 < line_idx < stop_line-1:
                    line_parts.append(html.Span(line, className="text-highlight secondary"))
                else:
                    line_parts.append(line)
                preview_components.extend(line_parts)
                preview_components.append(html.Br())
            # preview_components.append(html.Span("[...]", style={"color": "#888"}))

        pane_content = html.Pre(
            children=preview_components,
            id={"type": "preview-pre", "domain_uuid": domain_uuid},
            className="preview-pane",
            style={
                "width": "95%",
                "max-width": "600px",
                "height": "235px",
                "min-height": "235px",
                "fontFamily": "monospace",
                "fontWeight": "normal",
                "padding": "5px",
                "overflow": "auto",
                "whiteSpace": "pre-wrap",
                "resize": "vertical",
            }
        )
    else:
        if base_path == "" or param_target_file == "":
            text = f"Invalid architecture settings. Unable to preview."
        elif settings == {}:
            text = f"No settings found."
        else:
            text = f"Preview file '{os.path.realpath(param_target_file)}' not found. Unable to preview."
        pane_content = html.Div(text, className="error")
    return preview_div(pane_content)

def preview_div(content):
    return html.Div(
        children=[
            html.H3("Preview Pane"),
            content,
        ], 
    )

def domain_section(domain: str, arch_name: str = "", settings: dict = {}, domain_uuid=None):
    # Generate a unique UUID for non-main domains
    if not domain_uuid:
        domain_uuid = get_uuid() if domain != hard_settings.main_parameter_domain else domain
    
    # Hidden divs for better animations
    hidden_count = random.randint(4, 7)
    title_hidden_divs = [html.Div(style={"display": "none"}) for _ in range(hidden_count)]
    hidden_count = random.randint(8, 12)
    content_hidden_divs = [html.Div(style={"display": "none"}) for _ in range(hidden_count)]
    
    # Domain Section
    return html.Div(
        children=[
            html.Div(
                children=[
                    *title_hidden_divs,
                    parameter_domain_title(domain_name=domain, domain_uuid=domain_uuid, arch_name=arch_name)
                ], 
                id=f"param-domain-title-div-{domain_uuid}",
                className="card-matrix config",
                style={"marginLeft": "-13px"},
            ),
            html.Div(
                children=[
                    *content_hidden_divs,
                    html.Div(
                        children=[
                            config_parameters_form(domain_uuid, settings),
                        ],
                        id={"type": "config-parameters", "domain_uuid": domain_uuid}, 
                        className="tile config"),
                    html.Div(id={"type": "preview-pane", "domain_uuid": domain_uuid}, className="tile config"),
                ], 
                className="card-matrix config",
                style={"marginLeft": "-13px"},
            ),
            html.Div([
                html.Div(
                    children=[add_card(domain_uuid=domain_uuid)],
                    id={"type": "config-cards-row", "domain_uuid": domain_uuid},
                    className=f"card-matrix configs", 
                ),
            ]),
            dcc.Store(id={"type": "config-params-store", "domain_uuid": domain_uuid}, data=settings),
            dcc.Store(id={"type": "domain-metadata", "domain_uuid": domain_uuid}, data={"domain_name": domain, "domain_uuid": domain_uuid}),
        ],
        id = {"type": "param-domain-section", "domain_uuid": domain_uuid},
    )


######################################
# Callbacks
######################################

@dash.callback(
    Output(f"param-domain-title-div-{hard_settings.main_parameter_domain}", "children"),
    Input("param-domains-section", "children"),
    State("url", "search"),
    preview_initial_call=True
)
def update_main_domain_title(_, search):
    arch_name = get_key_from_url(search, "arch")
    if not arch_name:
        return "No architecture selected.", dash.no_update
    return parameter_domain_title(domain_uuid=hard_settings.main_parameter_domain, arch_name=arch_name)

@dash.callback(
    Output("param-domains-section", "children"),
    Output("param-domains-section-initialized", "data"),
    Output("param-domains-section-update", "data"),
    Input({"page": page_path, "type": "architecture-title-div"}, "children"), # Trigger when title is loaded
    State("url", "search"),
    State("url", "pathname"),
    Input({"action": "add-domain", "type": dash.ALL}, "n_clicks"),
    Input({"action": "duplicate-domain", "domain_uuid": dash.ALL}, "n_clicks"),
    Input({"action": "delete-domain", "domain_uuid": dash.ALL}, "n_clicks"),
    State("odatix-settings", "data"),
    State("param-domains-section", "children"),
    State({"type": "domain-metadata", "domain_uuid": dash.ALL}, "data"),
    State("param-domains-section-update", "data"),
    prevent_initial_call=True
)
def update_param_domains(
    _, search, page, add_domain_click, duplicate_domain_click, delete_domain_click,
    odatix_settings, domain_sections, metadata, update_flag
):
    triggered_id = ctx.triggered_id
    if triggered_id == "url":
        if page != page_path:
            return dash.no_update, dash.no_update, dash.no_update

    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    arch_name = get_key_from_url(search, "arch")
    if not arch_name:
        return html.Div(
            children=[
                html.Div("No architecture selected.", className="error")
            ],
            className="card-matrix config",
            style={"marginLeft": "-13px"},
        ), dash.no_update, dash.no_update

    add_domain_div = html.Div(
        children=[
            add_parameter_domain_button("Add new parameter domain")
        ],
        className="card-matrix config",
        style={"marginLeft": "-13px"},
    )

    if not workspace.architecture_exists(arch_path, arch_name):
        domain_sections = []
        domain_sections.append(domain_section(hard_settings.main_parameter_domain, arch_name, settings={}))
        domain_sections.append(add_domain_div)
        return domain_sections, dash.no_update, dash.no_update

    domains = workspace.get_param_domains(arch_path, arch_name)

    # Generate domain sections
    if triggered_id == {"page": page_path, "type": "architecture-title-div"}:
        domain_sections = []
        settings = workspace.load_architecture_settings(arch_path, arch_name, hard_settings.main_parameter_domain)
        domain_sections.append(domain_section(hard_settings.main_parameter_domain, settings=settings))
        for domain in domains:
            settings = workspace.load_architecture_settings(arch_path, arch_name, domain)
            domain_sections.append(domain_section(domain, arch_name, settings=settings))
        domain_sections.append(add_domain_div)
        return domain_sections, True, dash.no_update
    
    elif isinstance(triggered_id, dict):
        trigger_action = triggered_id.get("action", "")
        # Add a new domain
        if trigger_action == "add-domain":
            base_name = "new_domain"
            suffix = 1
            new_domain = f"{base_name}{suffix}"
            while new_domain in domains:
                suffix += 1
                new_domain = f"{base_name}{suffix}"
            workspace.create_parameter_domain(arch_path, arch_name, new_domain)
            
            # Insert new domain section before the add domain button
            domain_sections = domain_sections[:-1] if isinstance(domain_sections, list) else []
            domain_sections.append(domain_section(new_domain, arch_name, settings={}))
            domain_sections.append(add_domain_div)
            return domain_sections, dash.no_update, dash.no_update

        # Duplicate domain
        elif trigger_action == "duplicate-domain":
            domain_to_duplicate_uuid  = triggered_id.get("domain_uuid", "")

            # Check if button was actually clicked
            domain_to_duplicate = None
            domain_to_duplicate_idx = None
            for i, data in enumerate(metadata):
                domain_uuid = data.get("domain_uuid", "")
                domain_name = data.get("domain_name", "")
                if domain_uuid == domain_to_duplicate_uuid:
                    domain_to_duplicate = domain_name
                    domain_to_duplicate_idx = i - 1 # -1 to account for main domain
                    break

            if domain_to_duplicate_uuid :
                if domain_to_duplicate_uuid  == hard_settings.main_parameter_domain:
                    base_name = "main_copy"
                else:
                    base_name = f"{domain_to_duplicate}_copy"
                suffix = 1
                new_domain = f"{base_name}{suffix}"
                while new_domain in domains:
                    suffix += 1
                    new_domain = f"{base_name}{suffix}"
                workspace.duplicate_parameter_domain(
                    arch_path, arch_name, arch_name, domain_to_duplicate, new_domain
                )

                # Insert new domain section before the add domain button
                domain_sections = domain_sections[:-1] if isinstance(domain_sections, list) else []
                new_domain_settings = workspace.load_architecture_settings(arch_path, arch_name, new_domain)
                domain_uuid = get_uuid()
                domain_sections.append(domain_section(new_domain, arch_name, settings=new_domain_settings, domain_uuid=domain_uuid))
                domain_sections.append(add_domain_div)
                return domain_sections, dash.no_update, domain_uuid

        # Delete domain
        elif trigger_action == "delete-domain":
            domain_to_delete_uuid = triggered_id.get("domain_uuid", "")

            # Check if button was actually clicked
            domain_to_delete = None
            domain_to_delete_idx = None
            for i, data in enumerate(metadata):
                domain_uuid = data.get("domain_uuid", "")
                domain_name = data.get("domain_name", "")
                if domain_uuid == domain_to_delete_uuid:
                    domain_to_delete = domain_name
                    domain_to_delete_idx = i - 1 # -1 to account for main domain
                    break
            if domain_to_delete_idx and delete_domain_click[domain_to_delete_idx] == 0:
                return dash.no_update, dash.no_update, dash.no_update
            
            if domain_to_delete and domain_to_delete != hard_settings.main_parameter_domain:
                workspace.delete_parameter_domain(arch_path, arch_name, domain_to_delete)
                if isinstance(domain_sections, list):
                    for i, section in enumerate(domain_sections):
                        domain = section.get("props", {}).get("id", {}).get("domain_uuid", "")
                        if domain == domain_to_delete_uuid:
                            domain_sections.pop(i)
                            return domain_sections, dash.no_update, dash.no_update

    return dash.no_update, dash.no_update, dash.no_update

@dash.callback(
    Output({"type": "config-cards-row", "domain_uuid": dash.ALL}, "children"),
    State("url", "search"),
    Input("param-domains-section-initialized", "data"),
    Input("config-layout-dropdown", "value"),
    Input({"type": "new-config", "domain_uuid": dash.ALL}, "n_clicks"),
    Input({"type": "save-config", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "n_clicks"),
    Input({"type": "delete-config", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "n_clicks"),
    Input({"type": "duplicate-config", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "n_clicks"),
    Input("param-domains-section-update", "data"),
    State({"type": "config-cards-row", "domain_uuid": dash.ALL}, "children"),
    State({"type": "config-title", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "value"),
    State({"type": "config-content", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "value"),
    State({"type": "config-metadata", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "data"),
    State({"type": "domain-metadata", "domain_uuid": dash.ALL}, "data"),
    State("odatix-settings", "data"),
    prevent_initial_call=True
)
def update_config_cards(
    search, _,
    config_layout, add_click, save_clicks, delete_clicks, duplicate_clicks, update_domain_uuid,
    config_cards_row, title_values, contents, config_metadata, domain_metadata, odatix_settings
):
    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    arch_name = get_key_from_url(search, "arch")
    if not arch_name:
        return [html.Div("No architecture selected.", className="error")]

    triggered_id = ctx.triggered_id
    if not isinstance(triggered_id, dict):
        domains = {}
        for data in domain_metadata:
            domain_name = data.get("domain_name", "")
            domain_uuid = data.get("domain_uuid", "")
            domains[domain_uuid] = domain_name
    else:
        domains = {}
        trig_type = triggered_id.get("type", None)
        trig_domain_uuid = triggered_id.get("domain_uuid", hard_settings.main_parameter_domain)
        trig_domain_name = ""

        # Get domain from metadata
        trig_domain_idx = -1
        for i, data in enumerate(domain_metadata):
            domain_name = data.get("domain_name", "")
            domain_uuid = data.get("domain_uuid", "")
            domains[domain_uuid] = domain_name
            if domain_uuid == trig_domain_uuid:
                trig_domain_name = domain_name
                trig_domain_idx = i

        if trig_type in ["new-config", "save-config", "delete-config", "duplicate-config"]:         
            trig_domain_configs = []
            trig_config_uuid = triggered_id.get("config_uuid", "")
            trig_config_name = ""
            trig_config_initial_content = ""
            
            # Get config from metadata
            trig_global_idx = -1
            trig_config_idx = -1
            domain_config_idx = -1
            for i, data in enumerate(config_metadata):
                domain_uuid = data.get("domain_uuid", "")
                if domain_uuid == trig_domain_uuid:
                    domain_config_idx += 1
                    config_name = data.get("config_name", "")
                    config_uuid = data.get("config_uuid", "")
                    config_content = data.get("config_content", "")
                    trig_domain_configs.append(config_name)
                    if config_uuid == trig_config_uuid:
                        trig_config_name = config_name
                        trig_config_initial_content = config_content
                        trig_config_idx = domain_config_idx
                        trig_global_idx = i
            
            if trig_type == "new-config" and add_click:
                for idx in range(1, 1001):
                    new_filename = f"new_config{idx}.txt"
                    if new_filename not in trig_domain_configs:
                        workspace.save_config_file(arch_path, arch_name, trig_domain_name, new_filename, "")
                        config_uuid = get_uuid()
                        config_cards_row[trig_domain_idx].insert(-1, 
                            config_card(
                                domain_uuid=trig_domain_uuid,
                                config_uuid=config_uuid,
                                config_name=new_filename,
                                content="",
                                initial_content="",
                                config_layout=config_layout,
                            )
                        )
                        break
                else:
                    error_msg = "Too many config creation fails (1000 max)."

            elif trig_type in ["save-config", "delete-config", "duplicate-config"]:         
                trig_config_content = contents[trig_global_idx] if trig_global_idx >= 0 and trig_global_idx < len(contents) else ""
                
                # Save config (and handle rename)
                if trig_type == "save-config":
                    config_new_title = title_values[trig_global_idx] if trig_global_idx >= 0 and trig_global_idx < len(title_values) else ""
                    config_old_title = trig_config_name
                    
                    if not config_new_title.endswith(".txt"):
                        config_new_title = config_new_title + ".txt"
                    if config_new_title != config_old_title:
                        print(f"Renaming config from '{config_old_title}' to '{config_new_title}'")
                        if config_new_title in trig_domain_configs:
                            if verbose:
                                print(f"File '{config_new_title}' already exists.")
                        else:
                            path = workspace.get_arch_domain_path(arch_path, arch_name, trig_domain_name)
                            old_path = os.path.join(path, config_old_title)
                            new_path = os.path.join(path, config_new_title)
                            if verbose:
                                print(f"Renaming {old_path} to {new_path}")
                            os.rename(old_path, new_path)
                            config_old_title = config_new_title
                    if verbose:
                        print(f"Saving config '{config_old_title}' in domain '{trig_domain_name}'")
                    workspace.save_config_file(arch_path, arch_name, trig_domain_name, config_new_title, trig_config_content)
                    config_metadata[trig_config_idx]['config_content'] = trig_config_content
                    config_cards_row[trig_domain_idx][trig_config_idx] = config_card(
                        domain_uuid=trig_domain_uuid,
                        config_uuid=trig_config_uuid,
                        config_name=config_old_title,
                        content=trig_config_content,
                        initial_content=trig_config_content,
                        config_layout=config_layout,
                    )
                
                # Delete config
                if trig_type == "delete-config":
                    workspace.delete_config_file(arch_path, arch_name, trig_domain_name, trig_config_name)
                    config_cards_row[trig_domain_idx].pop(trig_config_idx)

                # Duplicate config
                if trig_type == "duplicate-config":
                    base = trig_config_name[:-4] if trig_config_name.endswith(".txt") else trig_config_name
                    suffix = 1
                    new_filename = f"{base}_copy{suffix}.txt"
                    while new_filename in trig_domain_configs:
                        suffix += 1
                        new_filename = f"{base}_copy{suffix}.txt"
                    workspace.save_config_file(arch_path, arch_name, trig_domain_name, new_filename, trig_config_content)
                    config_uuid = get_uuid()
                    config_cards_row[trig_domain_idx].insert(-1, 
                        config_card(
                            domain_uuid=trig_domain_uuid,
                            config_uuid=config_uuid,
                            config_name=new_filename,
                            content=trig_config_content,
                            initial_content=trig_config_content,
                            config_layout=config_layout,
                        )
                    )
  
    # Initial load
    if triggered_id == "param-domains-section-initialized":
        config_cards_row = []
        for idx, (domain_uuid, domain_name) in enumerate(domains.items()):
            files = workspace.get_config_files(arch_path, arch_name, domain_name)
            config_cards = []
            config_metadata = []
            for config_name in files:
                config_content = workspace.load_config_file(arch_path, arch_name, domain_name, config_name)
                config_uuid = get_uuid()
                # Create config card
                config_cards.append(
                    config_card(
                        domain_uuid = domain_uuid, 
                        config_uuid = config_uuid,
                        config_name = config_name,
                        content = config_content,
                        initial_content = config_content,
                        config_layout = config_layout,
                    )
                )
            config_cards.append(add_card(domain_uuid=domain_uuid))
            config_cards_row.append(config_cards)
        return config_cards_row
    
    # If a domain was duplicated, load the new domain configs
    elif triggered_id == "param-domains-section-update":
        # Reload only domain with uuid == update_domain_uuid
        for i, domain_uuid in enumerate(domains.keys()):
            if domain_uuid == update_domain_uuid:
                domain_name = domains[domain_uuid]
                files = workspace.get_config_files(arch_path, arch_name, domain_name)
                config_cards = []
                for config_name in files:
                    config_content = workspace.load_config_file(arch_path, arch_name, domain_name, config_name)
                    config_uuid = get_uuid()
                    # Create config card
                    config_cards.append(
                        config_card(
                            domain_uuid = domain_uuid, 
                            config_uuid = config_uuid,
                            config_name = config_name,
                            content = config_content,
                            initial_content = config_content,
                            config_layout = config_layout,
                        )
                    )
                config_cards.append(add_card(domain_uuid=domain_uuid))
                config_cards_row[i] = config_cards
                break

    return config_cards_row

@dash.callback(
    Output({"type": "preview-pane", "domain_uuid": dash.ALL}, "children"),
    State("url", "search"),
    Input("param-domains-section", "children"),
    Input({"type": "config-cards-row", "domain_uuid": dash.ALL}, "children"),
    Input({"type": "use_parameters", "domain_uuid": dash.ALL}, "value"),
    Input({"type": "param_target_file", "domain_uuid": dash.ALL}, "value"),
    Input({"type": "start_delimiter", "domain_uuid": dash.ALL}, "value"),
    Input({"type": "stop_delimiter", "domain_uuid": dash.ALL}, "value"),
    Input({"type": "config-params-store", "domain_uuid": dash.ALL}, "data"),
    Input({"type": "config-content", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "value"),
    State({"type": "config-metadata", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "data"),
    State({"type": "domain-metadata", "domain_uuid": dash.ALL}, "data"),
    State("odatix-settings", "data"),
    prevent_initial_call=True
)
def update_preview_all(
    search, param_domains_update,
    config_cards_rows, params_enables, target_files, start_delims, stop_delims, settings_list, config_contents_list, 
    config_metadata, domain_metadata, odatix_settings
):
    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    arch_name = get_key_from_url(search, "arch")
    if not workspace.architecture_exists(arch_path, arch_name):
        return [
            html.Div([
                html.Div("Architecture is not created yet."),
                html.Div("Edit settings, then save or add a new config to create it."),
            ], className="error warning")
        ] * len(config_cards_rows)

    # Group config contents by domain
    contents_by_domain = []
    domain_contents = []
    for domain in domain_metadata:
        domain_uuid = domain.get("domain_uuid", "")
        indices = [i for i, config in enumerate(config_metadata) if config.get("domain_uuid", "") == domain_uuid]
        domain_contents = [config_contents_list[i] for i in indices]
        contents_by_domain.append(domain_contents)

    # Generate previews for each domain
    results = []
    for i, domain in enumerate(domain_metadata):
        domain_uuid = domain.get("domain_uuid", "")
        config_contents_list = contents_by_domain[i] if i < len(contents_by_domain) else []
        settings = settings_list[0] if 0 < len(settings_list) and settings_list[0] is not None else {}
        domain_settings = settings_list[i] if i < len(settings_list) and settings_list[i] is not None else {}
        domain_settings["use_parameters"] = params_enables[i] if i < len(params_enables) else ""
        domain_settings["param_target_file"] = target_files[i] if i < len(target_files) else ""
        domain_settings["start_delimiter"] = start_delims[i] if i < len(start_delims) else ""
        domain_settings["stop_delimiter"] = stop_delims[i] if i < len(stop_delims) else ""
        replacement_text = config_contents_list[0] if config_contents_list else ""
        results.append(preview_pane(domain_uuid, settings, domain_settings, replacement_text))
    return results

@dash.callback(
    Output({"type": "save-config", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "className"),
    Output({"type": "save-config", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "data-tooltip"),
    Input("param-domains-section", "children"),
    Input({"type": "config-title", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "value"),
    Input({"type": "config-content", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "value"),
    Input({"type": "initial-title", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "data"),
    Input({"type": "initial-content", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "data"),
)
def update_save_status(param_domains_section, title_values, content_values, initial_titles, initial_contents):
    save_classes = []
    tooltip_texts = []
    for title, content, initial_title, initial_content in zip(title_values, content_values, initial_titles, initial_contents):
        # Check for empty title
        if not title:
            save_classes.append("color-button error-status icon-button tooltip delay bottom small")
            tooltip_texts.append("Configuration name cannot be empty")
            continue
        # Check for invalid characters in title
        invalid_char_found = False
        for c in hard_settings.invalid_filename_characters:
            if c in title:
                c = "' ' (space)" if c == " " else f"'{c}'"
                save_classes.append("color-button error-status icon-button tooltip delay bottom small")
                tooltip_texts.append(f"Unauthorized character in configuration name: {c}")
                invalid_char_found = True
                break
        if invalid_char_found:
            continue
        # Check if a config with the same name already exists in the domain
        if title in initial_titles and title != initial_title:
            save_classes.append("color-button error-status icon-button tooltip delay bottom small")
            tooltip_texts.append("A configuration with this name already exists in the domain.")
            continue
        # Check for changes
        if title != initial_title or content != initial_content:
            save_classes.append("color-button warning icon-button tooltip delay bottom small")
            tooltip_texts.append("Unsaved changes!")
        else:
            save_classes.append("color-button disabled icon-button tooltip delay bottom small")
            tooltip_texts.append("Nothing to save")
    return save_classes, tooltip_texts

@dash.callback(
    Output({"type": "save-params-btn", "domain_uuid": dash.ALL}, "className"),
    Input({"type": "use_parameters", "domain_uuid": dash.ALL}, "value"),
    Input({"type": "param_target_file", "domain_uuid": dash.ALL}, "value"),
    Input({"type": "start_delimiter", "domain_uuid": dash.ALL}, "value"),
    Input({"type": "stop_delimiter", "domain_uuid": dash.ALL}, "value"),
    Input({"type": "config-params-store", "domain_uuid": dash.ALL}, "data"),
)
def update_params_save_button(params_enables, target_files, start_delims, stop_delims, settings_list):
    save_classes = []
    disabled_class = "color-button disabled icon-button"
    enabled_class = "color-button warning icon-button"

    for use_parameters, param_target_file, start_delimiter, stop_delimiter, domain_settings in zip(params_enables, target_files, start_delims, stop_delims, settings_list):
        if domain_settings is None:
            domain_settings = {}
        use_parameters = True if use_parameters else False # Convert from [True]/[] to True/False
        if use_parameters != domain_settings.get("use_parameters", True):
            save_classes.append(enabled_class)
        elif param_target_file != domain_settings.get("param_target_file", ""):
            save_classes.append(enabled_class)
        elif start_delimiter != domain_settings.get("start_delimiter", ""):
            save_classes.append(enabled_class)
        elif stop_delimiter != domain_settings.get("stop_delimiter", ""):
            save_classes.append(enabled_class)
        else:
            save_classes.append(disabled_class)
    return save_classes

@dash.callback(
    Output({"type": "config-card", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "className"),
    Output({"type": "add-config-card", "domain_uuid": dash.ALL}, "className"),
    Output({"type": "config-cards-row", "domain_uuid": dash.ALL}, "className"),
    Input("config-layout-dropdown", "value"),
    State({"type": "config-card", "domain_uuid": dash.ALL, "config_uuid": dash.ALL}, "className"),
    State({"type": "add-config-card", "domain_uuid": dash.ALL}, "className"),
    State({"type": "config-cards-row", "domain_uuid": dash.ALL}, "className"),
)
def update_layout_style(layout_value, config_card_classes, add_card_classes, config_row_classes):
    if layout_value == "wide":
        config_card_classes = ["card configs wide" for _ in range(len(config_card_classes))]
        add_card_classes = ["card configs add wide hover" for _ in range(len(add_card_classes))]
        config_row_classes = ["card-matrix configs wide" for _ in range(len(config_row_classes))]
    else:
        config_card_classes = ["card configs" for _ in range(len(config_card_classes))]
        add_card_classes = ["card configs add hover" for _ in range(len(add_card_classes))]
        config_row_classes = ["card-matrix configs" for _ in range(len(config_row_classes))]
    return config_card_classes, add_card_classes, config_row_classes

@dash.callback(
    Output({"type": "config-params-store", "domain_uuid": dash.ALL}, "data"),
    Input({"type": "save-params-btn", "domain_uuid": dash.ALL}, "n_clicks"),
    State({"type": "use_parameters", "domain_uuid": dash.ALL}, "value"),
    State({"type": "param_target_file", "domain_uuid": dash.ALL}, "value"),
    State({"type": "start_delimiter", "domain_uuid": dash.ALL}, "value"),
    State({"type": "stop_delimiter", "domain_uuid": dash.ALL}, "value"),
    State({"type": "domain-metadata", "domain_uuid": dash.ALL}, "data"),
    State({"type": "config-params-store", "domain_uuid": dash.ALL}, "data"),
    State("url", "search"),
    State("odatix-settings", "data"),
)
def save_config_parameters(
    n_clicks, 
    use_parameters, param_target_files, start_delimiters, stop_delimiters, domain_metadata, config_params_stores,
    search, odatix_settings
):
    arch_name = get_key_from_url(search, "arch")
    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)

    triggered = ctx.triggered_id
    if isinstance(triggered, dict):        
        trig_domain_uuid = triggered.get("domain_uuid", hard_settings.main_parameter_domain)
        trig_domain_name = trig_domain_uuid
        
        # Get domain from metadata
        for data in domain_metadata:
            domain_name = data.get("domain_name", "")
            domain_uuid = data.get("domain_uuid", "")
            if domain_uuid == trig_domain_uuid :
                trig_domain_name = domain_name
                break

        idx = next((i for i, data in enumerate(domain_metadata) if data.get("domain_uuid") == trig_domain_uuid), -1)
        if idx != -1:
            settings = config_params_stores[idx] if idx < len(config_params_stores) and config_params_stores[idx] is not None else {}

            # Get values
            use_parameters = use_parameters[idx] if idx < len(use_parameters) else False
            use_parameters = True if use_parameters else False # Convert from [True]/[] to True/False
            param_target_file = param_target_files[idx] if idx < len(param_target_files) else ""
            start_delimiter = start_delimiters[idx] if idx < len(start_delimiters) else ""
            stop_delimiter = stop_delimiters[idx] if idx < len(stop_delimiters) else ""

            # Update settings
            settings["use_parameters"] = use_parameters
            settings["param_target_file"] = param_target_file
            settings["start_delimiter"] = start_delimiter
            settings["stop_delimiter"] = stop_delimiter

            workspace.update_domain_settings(arch_path, arch_name, trig_domain_name, settings)
            return [settings if i == idx else dash.no_update for i in range(len(domain_metadata))]
    return [dash.no_update for _ in range(len(domain_metadata))]

@dash.callback(
    Output({"type": "params-config-fields", "domain_uuid": dash.ALL}, "className"),
    Output({"type": "config-cards-row", "domain_uuid": dash.ALL}, "style"),
    Input ({"type": "use_parameters", "domain_uuid": dash.ALL}, "value"),
)
def toggle_params_fields(enabled_values):
    styles = []
    classes = []
    for value in enabled_values:
        if value:
            styles.append({})
            classes.append("animated-section")
        else:
            styles.append(Style.hidden)
            classes.append("animated-section hide")
    return classes, styles

@dash.callback(
    Output({"page": page_path, "type": "architecture-title-div"}, "children"),
    Input("url", "search"),
)
def update_architecture_title(search):
    arch_name = get_key_from_url(search, "arch")
    if not arch_name:
        arch_name = "New_Architecture"
    return architecture_title(arch_name)

@dash.callback(
    Output({"type": "save-domain-title", "domain_uuid": dash.ALL}, "className"),
    Output({"type": "save-domain-title", "domain_uuid": dash.ALL}, "data-tooltip"),
    Output({"type": "domain-metadata", "domain_uuid": dash.ALL}, "data"),
    Output({"type": "generate-config", "domain_uuid": dash.ALL, "is_link": True}, "href"),
    Input({"type": "save-domain-title", "domain_uuid": dash.ALL}, "n_clicks"),
    Input({"type": "domain-title-input", "domain_uuid": dash.ALL}, "value"),
    State({"type": "domain-metadata", "domain_uuid": dash.ALL}, "data"),
    State("url", "search"),
    State("odatix-settings", "data"),
)
def update_params_title_save_button(_, title_input, domain_metadata, search, odatix_settings):
    save_classes = []
    tooltips = []
    new_metadata = []
    new_hrefs = []

    disabled_class = "color-button invisible icon-button tooltip delay bottom small"
    enabled_class = "color-button warning icon-button tooltip bottom small"
    error_class = "color-button disabled icon-button error-status tooltip bottom small"

    triggered_id = ctx.triggered_id if isinstance(ctx.triggered_id, dict) else {}
    triggered_type = triggered_id.get("type", "")
    triggered_domain_uuid = triggered_id.get("domain_uuid", "")
    
    arch_path = odatix_settings.get("arch_path", OdatixSettings.DEFAULT_ARCH_PATH)
    arch_name = get_key_from_url(search, "arch")
    if not arch_name:
        raise dash.exceptions.PreventUpdate

    domain_metadata = domain_metadata[1:] # Remove main domain entry
    new_metadata.append(dash.no_update)  # Placeholder for main domain
    for new_domain_name, metadata in zip(title_input, domain_metadata):
        domain_name = metadata.get("domain_name", "")
        domain_uuid = metadata.get("domain_uuid", "")

        # Save button clicked for this domain
        if triggered_type == "save-domain-title" and domain_uuid == triggered_domain_uuid:
            if new_domain_name != domain_name and new_domain_name != "":
                workspace.rename_parameter_domain(arch_path, arch_name, domain_name, new_domain_name)

                domain_name = new_domain_name
                save_classes.append(disabled_class)
                tooltips.append("Nothing to save")
                metadata["domain_name"] = new_domain_name
                new_metadata.append(metadata)
                new_link = generate_config_link(arch_name, new_domain_name)
                new_hrefs.append(new_link)
            else:
                new_metadata.append(dash.no_update)
                new_hrefs.append(dash.no_update)
                save_classes.append(dash.no_update)
                tooltips.append(dash.no_update)

        # No save button clicked, just check for changes
        else:
            new_metadata.append(dash.no_update)
            new_hrefs.append(dash.no_update)
            if new_domain_name != domain_name:
                invalid_char_found = False
                for c in hard_settings.invalid_filename_characters:
                    if c in new_domain_name:
                        c = "' ' (space)" if c == " " else f"'{c}'"
                        invalid_char_found = True
                        save_classes.append(error_class)
                        tooltips.append(f"Unauthorized character in parameter domain name: {c}")
                        break
                if invalid_char_found:  
                    continue
                if new_domain_name == "" or " " in new_domain_name:
                    save_classes.append(error_class)
                    tooltips.append("Parameter domain name cannot be empty")
                elif workspace.parameter_domain_exists(arch_path, arch_name, new_domain_name):
                    save_classes.append(error_class)
                    tooltips.append(f"Parameter domain '{new_domain_name}' already exists for this architecture")
                else:
                    save_classes.append(enabled_class)
                    tooltips.append("Save new parameter domain name")
            else:
                save_classes.append(disabled_class)
                tooltips.append("Nothing to save")
    return save_classes, tooltips, new_metadata, new_hrefs


######################################
# Layout
######################################

layout = html.Div(
    children=[
        dcc.Location(id="url"),
        html.Div(id={"page": page_path, "type": "architecture-title-div"}, style={"marginTop": "20px"}),
        html.Div(id="param-domains-section", style={"marginBottom": "10px"}),
        dcc.Store(id="param-domains-section-initialized", data=False),
        dcc.Store(id="param-domains-section-update", data=""),
    ],
    className="page-content",
    style={
        "padding": "0 16%",
        "display": "flex",  
        "flexDirection": "column",
        "min-height": f"calc(100vh - {navigation.top_bar_height})",
    },
)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import plotly.graph_objects as go
import plotly.io as pio
import copy

default_theme = "plotly"
templates = pio.templates

themes = {
  "code_dark": {
    "base_template": "plotly_dark",
    "nav_bgcolor": "#1e1e1e",
    "page_bgcolor": "#252526",
    "plot_bgcolor": "#252526",
    "button_color":" #aaaaaa",
    "button_active_color": "#ffffff",
  },
  "odatix_darker": { 
    "base_template": "plotly_dark",
    "nav_bgcolor":"#0b0b0b",
    "page_bgcolor": "#24292e",
    "plot_bgcolor": "#181b20",
    "button_color":" #aaaaaa",
    "button_active_color": "#ffffff",
  },
  "odatix_dark": { 
    "nav_bgcolor":"#181b20",
    "base_template": "plotly_dark",
    "page_bgcolor": "#24292e",
    "plot_bgcolor": "#24292e",
    "button_color":" #aaaaaa",
    "button_active_color": "#ffffff",
  },
  "odatix_light": { 
    "base_template": "plotly_white",
    "nav_bgcolor": "#e5ecf6",
    "page_bgcolor": "#ffffff",
    "plot_bgcolor": "#ffffff",
    "button_color":" #aaaaaa",
    "button_active_color": "#ffffff",
    "legend_default_color": "#24292e",
  },
  "plotly_dark": {
    "base_template": None,
    "nav_bgcolor":"#0b0b0b",
    "page_bgcolor": "#111111",
    "plot_bgcolor": "#111111",
    "button_color":" #aaaaaa",
    "button_active_color": "#ffffff",
  },
  "candy": {
    "base_template": "odatix_darker",
    "nav_bgcolor":"#E156E8",
    "page_bgcolor": "#954DC5",
    "plot_bgcolor": "#6D13A8",
    "button_color":"#DFA2FC",
    "button_active_color": "#FCA2E7",
  },
  "default": {
    "base_template": None,
    "nav_bgcolor":"#24292e",
    "page_bgcolor": "#ffffff",
    "plot_bgcolor": "#ffffff",
    "button_color":" #aaaaaa",
    "button_active_color": "#ffffff",
    "legend_default_color": "#ffffff",
  },
}

def get_page_bgcolor(theme, default=themes["default"]["page_bgcolor"]):
  if theme in themes and "page_bgcolor" in themes[theme]:
    return themes[theme]["page_bgcolor"]
  else:
    return default

def get_nav_bgcolor(theme, default=themes["default"]["nav_bgcolor"]):
  if theme in themes and "nav_bgcolor" in themes[theme]:
    return themes[theme]["nav_bgcolor"]
  else:
    return default

def get_plot_bgcolor(theme, default=themes["default"]["plot_bgcolor"]):
  if theme in themes and "plot_bgcolor" in themes[theme]:
    return themes[theme]["plot_bgcolor"]
  else:
    return default

def get_button_color(theme, default=themes["default"]["button_color"]):
  if theme in themes and "button_color" in themes[theme]:
    return themes[theme]["button_color"]
  else:
    return default

def get_button_active_color(theme, default=themes["default"]["button_active_color"]):
  if theme in themes and "button_active_color" in themes[theme]:
    return themes[theme]["button_active_color"]
  else:
    return default

def get_legend_default_color(theme, default=themes["default"]["legend_default_color"]):
  if theme in themes and "legend_default_color" in themes[theme]:
    return themes[theme]["legend_default_color"]
  else:
    return default

templates["candy"] = go.layout.Template(
  layout = copy.deepcopy(templates["plotly"].layout).update(
    paper_bgcolor=get_page_bgcolor("candy"),
    plot_bgcolor=get_plot_bgcolor("candy"),
    polar_bgcolor=get_plot_bgcolor("candy"),
    barcornerradius=10000,
  )
)

templates["code_dark"] = go.layout.Template(
  layout = copy.deepcopy(templates["plotly_dark"].layout).update(
    paper_bgcolor=get_page_bgcolor("code_dark"),
    plot_bgcolor=get_plot_bgcolor("code_dark"),
    polar_bgcolor=get_plot_bgcolor("code_dark"),
  )
)

templates["odatix_darker"] = go.layout.Template(
  layout = copy.deepcopy(templates["plotly_dark"].layout).update(
    paper_bgcolor=get_page_bgcolor("odatix_darker"),
    plot_bgcolor=get_plot_bgcolor("odatix_darker"),
    polar_bgcolor=get_plot_bgcolor("odatix_darker"),
  )
)

templates["odatix_dark"] = go.layout.Template(
  layout = copy.deepcopy(templates["plotly_dark"].layout).update(
    paper_bgcolor=get_page_bgcolor("odatix_dark"),
    plot_bgcolor=get_plot_bgcolor("odatix_dark"),
    polar_bgcolor=get_plot_bgcolor("odatix_dark"),
  )
)

templates["odatix_light"] = go.layout.Template(
  layout = copy.deepcopy(templates["plotly_white"].layout).update(
    paper_bgcolor=get_page_bgcolor("odatix_light"),
    plot_bgcolor=get_plot_bgcolor("odatix_light"),
    polar_bgcolor=get_plot_bgcolor("odatix_light"),
  )
)



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import sys
import yaml

from odatix.lib.get_from_dict import get_from_dict, Key, KeyNotInDictError, BadValueInDictError
import odatix.lib.printc as printc

script_name = os.path.basename(__file__)

def read_tool_settings(tool, tool_settings_file, synth_type='fmax_synthesis'):
  """
  Reads the settings for a given EDA tool from a YAML configuration file.

  Args:
    tool (str): The name of the EDA tool.
    tool_settings_file (str): The path to the YAML configuration file.
    synth_type (str, optional): The type of synthesis to be used. 
                                Can be either 'fmax_synthesis' or 'custom_freq_synthesis'. 
                                Defaults to 'fmax_synthesis'.

  Returns:
    tuple: A tuple containing:
      - process_group (bool): Whether the process grouping is enabled.
      - default_metrics_file (str): Path to the default metrics file.
      - synthesis_command (list): The command for the selected synthesis type.
      - tool_test_command (list): The command to check if the tool is installed.

  Raises:
    SystemExit: If the settings file does not exist, is invalid, or if the requested tool 
                or synthesis type is not supported.
  """
  if not os.path.isfile(tool_settings_file):
    printc.error(
      f'Settings file "{tool_settings_file}", for the selected EDA tool "{tool}" does not exist', script_name
    )
    sys.exit(-1)

  # Load YAML settings file
  with open(tool_settings_file, "r") as f:
    try:
      settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
    except Exception as e:
      printc.error(f'Settings file "{tool_settings_file}", for the selected EDA tool "{tool}" is not a valid YAML file', script_name)
      printc.cyan("Error details: ", end="", script_name=script_name)
      print(str(e))
      sys.exit(-1)

  # Retrieve global settings
  process_group, _ = get_from_dict("process_group", settings_data, tool_settings_file, default_value=True, script_name=script_name)
  report_path, _ = get_from_dict("report_path", settings_data, tool_settings_file, default_value="report", silent=True, script_name=script_name)
  default_metrics_file, _ = get_from_dict("default_metrics_file", settings_data, tool_settings_file, behavior=Key.MANTADORY, script_name=script_name)

  # Determine the current platform (Unix or Windows)
  platform_key = "windows" if sys.platform == 'win32' else "unix"

  # Ensure the corresponding section exists in the configuration file
  if platform_key not in settings_data:
    printc.error(f'The selected EDA tool "{tool}" does not support {platform_key}', script_name)
    sys.exit(-1)

  platform_settings = settings_data[platform_key]

  # Retrieve tool test command
  tool_test_command, tool_test_supported = get_from_dict("tool_test_command", platform_settings, tool_settings_file, silent=True, script_name=script_name)

  if not tool_test_supported:
    printc.error(f'The selected EDA tool "{tool}" is not supported on {platform_key} (tool_test_command is missing)', script_name)
    sys.exit(-1)

  # Retrieve synthesis command based on the requested type
  synthesis_key = "fmax_synthesis_command" if synth_type == 'fmax_synthesis' else "custom_freq_synthesis_command"
  synthesis_command, synthesis_supported = get_from_dict(synthesis_key, platform_settings, tool_settings_file, silent=True, script_name=script_name)

  if not synthesis_supported:
    printc.error(f'{synth_type.replace("_", " ").capitalize()} is not supported with the selected EDA tool "{tool}" on {platform_key}', script_name)
    sys.exit(-1)

  return process_group, report_path, synthesis_command, tool_test_command, default_metrics_file



# ********************************************************************** #
#                                Odatix                                  #
# ********************************************************************** #
#
# Copyright (C) 2022 Jonathan Saussereau
#
# This file is part of Odatix.
# Odatix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Odatix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Odatix. If not, see <https://www.gnu.org/licenses/>.
#

import os
import re
import sys
import math
import yaml
import copy
import itertools

from os.path import isfile
from os.path import isdir

from odatix.lib import hard_settings
from odatix.lib.settings import OdatixSettings
from odatix.lib.utils import *
from odatix.lib.get_from_dict import get_from_dict, Key, KeyNotInDictError, BadValueInDictError
import odatix.lib.printc as printc
from odatix.lib.variables import replace_variables, Variables
from odatix.lib.param_domain import ParamDomain

script_name = os.path.basename(__file__)

class Architecture:
    def __init__(
        self, arch_name, arch_display_name, lib_name, target, local_rtl_path, tmp_script_path, tmp_report_path, tmp_log_path, tmp_dir, 
        design_path, design_path_whitelist, design_path_blacklist, rtl_path, log_path, arch_path,
        clock_signal, reset_signal, top_level_module, top_level_filename, use_parameters, start_delimiter, stop_delimiter,
        file_copy_enable, file_copy_source, file_copy_dest, script_copy_enable, script_copy_source, 
        fmax_lower_bound, fmax_upper_bound, range_list, target_frequency,
        param_target_filename, generate_rtl, generate_command, constraint_filename, install_path, 
        param_domains, continue_on_error=False, force_single_thread=False,
    ):
        self.arch_name = arch_name
        self.arch_display_name = arch_display_name
        self.lib_name = lib_name
        self.target = target
        self.local_rtl_path = local_rtl_path
        self.tmp_script_path = tmp_script_path
        self.tmp_report_path = tmp_report_path
        self.tmp_log_path = tmp_log_path
        self.tmp_dir = tmp_dir
        self.design_path = design_path
        self.design_path_whitelist = design_path_whitelist
        self.design_path_blacklist = design_path_blacklist
        self.rtl_path = rtl_path
        self.log_path = log_path
        self.arch_path = arch_path
        self.clock_signal = clock_signal
        self.reset_signal = reset_signal
        self.top_level_module = top_level_module
        self.top_level_filename = top_level_filename
        self.file_copy_enable = file_copy_enable
        self.file_copy_source = file_copy_source
        self.file_copy_dest = file_copy_dest
        self.script_copy_enable = script_copy_enable
        self.script_copy_source = script_copy_source
        self.fmax_lower_bound = fmax_lower_bound
        self.fmax_upper_bound = fmax_upper_bound
        self.range_list = range_list
        self.target_frequency = target_frequency
        self.param_target_filename = param_target_filename
        self.use_parameters = use_parameters
        self.start_delimiter = start_delimiter
        self.stop_delimiter = stop_delimiter
        self.generate_rtl = generate_rtl
        self.generate_command = generate_command
        self.constraint_filename = constraint_filename
        self.install_path = install_path
        self.param_domains = param_domains
        self.continue_on_error = continue_on_error
        self.force_single_thread = force_single_thread

    def write_yaml(arch, config_file): 
        domain_dict=dict()
        for param_domain in arch.param_domains:
            domain_dict[param_domain.domain] = param_domain.domain_value
        yaml_data = {
            'arch_name': arch.arch_name,
            'arch_display_name': arch.arch_display_name,
            'lib_name': arch.lib_name,
            'target': arch.target,
            'rtl_path': arch.local_rtl_path,
            'script_path': arch.tmp_script_path,
            'report_path': arch.tmp_report_path,
            'log_path': arch.tmp_log_path,
            'tmp_path': arch.tmp_dir,
            'design_path': arch.design_path,
            'design_path_whitelist': arch.design_path_whitelist,
            'design_path_blacklist': arch.design_path_blacklist,
            'source_rtl_path': arch.rtl_path,
            'arch_path': arch.arch_path,
            'clock_signal': arch.clock_signal,
            'reset_signal': arch.reset_signal,
            'top_level_module': arch.top_level_module,
            'top_level_file': arch.top_level_filename,
            'use_parameters': arch.use_parameters,
            'start_delimiter': arch.start_delimiter,
            'stop_delimiter': arch.stop_delimiter,
            'file_copy_enable': arch.file_copy_enable,
            'file_copy_source': arch.file_copy_source,
            'file_copy_dest': arch.file_copy_dest,
            'script_copy_enable': arch.script_copy_enable,
            'script_copy_source': arch.script_copy_source,
            'fmax_lower_bound': arch.fmax_lower_bound,
            'fmax_upper_bound': arch.fmax_upper_bound,
            'range_list': arch.range_list,
            'target_frequency': arch.target_frequency,
            'param_target_filename': arch.param_target_filename,
            'generate_rtl': arch.generate_rtl,
            'generate_command': arch.generate_command,
            'constraint_filename': arch.constraint_filename,
            'install_path': arch.install_path,
            'param_domains': domain_dict,
            'continue_on_error': arch.continue_on_error,
            'force_single_thread': arch.force_single_thread,
        }
            
        with open(config_file, 'w') as f:
            yaml.dump(yaml_data, f, default_flow_style=False, sort_keys=False)

    def read_yaml(config_file):
        if not os.path.isfile(config_file):
            printc.error("Settings file \"" + config_file + "\" does not exist", script_name)
            return None

        with open(config_file, 'r') as f:
            yaml_data = yaml.safe_load(f)
        
        try:
            arch = Architecture(
                arch_name                = get_from_dict("arch_name", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                arch_display_name        = get_from_dict("arch_display_name", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                lib_name                 = get_from_dict("lib_name", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                target                   = get_from_dict("target", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                rtl_path                 = get_from_dict("source_rtl_path", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                tmp_script_path          = get_from_dict("script_path", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                tmp_report_path          = get_from_dict("report_path", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                tmp_log_path             = get_from_dict("log_path", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                log_path                 = get_from_dict("log_path", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                tmp_dir                  = get_from_dict("tmp_path", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                design_path              = get_from_dict("design_path", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                design_path_whitelist    = get_from_dict("design_path_whitelist", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                design_path_blacklist    = get_from_dict("design_path_blacklist", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                local_rtl_path           = get_from_dict("rtl_path", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                arch_path                = get_from_dict("arch_path", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                clock_signal             = get_from_dict("clock_signal", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                reset_signal             = get_from_dict("reset_signal", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                top_level_module         = get_from_dict("top_level_module", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                top_level_filename       = get_from_dict("top_level_file", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                use_parameters           = get_from_dict("use_parameters", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                start_delimiter          = get_from_dict("start_delimiter", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                stop_delimiter           = get_from_dict("stop_delimiter", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                file_copy_enable         = get_from_dict("file_copy_enable", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                file_copy_source         = get_from_dict("file_copy_source", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                file_copy_dest           = get_from_dict("file_copy_dest", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                script_copy_enable       = get_from_dict("script_copy_enable", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                script_copy_source       = get_from_dict("script_copy_source", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                fmax_lower_bound         = get_from_dict("fmax_lower_bound", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                fmax_upper_bound         = get_from_dict("fmax_upper_bound", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                range_list               = get_from_dict("range_list", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                target_frequency         = get_from_dict("target_frequency", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                param_target_filename    = get_from_dict("param_target_filename", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                generate_rtl             = get_from_dict("generate_rtl", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                generate_command         = get_from_dict("generate_command", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                constraint_filename      = get_from_dict("constraint_filename", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],   
                install_path             = get_from_dict("install_path", yaml_data, config_file, behavior=Key.MANTADORY_RAISE, script_name=script_name)[0],
                param_domains            = get_from_dict("install_path", yaml_data, config_file, default_value={}, script_name=script_name)[0],
                continue_on_error        = get_from_dict("continue_on_error", yaml_data, config_file, default_value=False, script_name=script_name)[0],
                force_single_thread      = get_from_dict("force_single_thread", yaml_data, config_file, default_value=False, script_name=script_name)[0],
            )
        except (KeyNotInListError, BadValueInListError):
            return None
        return arch

class ArchitectureHandler:

    def __init__(
        self,
        work_path,
        arch_path,
        script_path,
        log_path,
        work_rtl_path,
        work_script_path,
        work_log_path,
        work_report_path,
        process_group,
        command,
        eda_target_filename,
        fmax_status_filename,
        frequency_search_filename,
        param_settings_filename,
        valid_status,
        valid_frequency_search,
        forced_fmax_lower_bound,
        forced_fmax_upper_bound,
        forced_custom_freq_list,
        overwrite,
        continue_on_error=False,
        force_single_thread=False,
    ):
        self.work_path = work_path
        self.arch_path = arch_path
        self.script_path = script_path
        self.log_path = log_path
        self.work_rtl_path = work_rtl_path
        self.work_script_path = work_script_path
        self.work_log_path = work_log_path
        self.work_report_path = work_report_path

        self.process_group = process_group
        self.command = command
        
        self.eda_target_filename = eda_target_filename
        self.fmax_status_filename = fmax_status_filename
        self.frequency_search_filename = frequency_search_filename
        self.param_settings_filename = param_settings_filename
        
        self.valid_status = valid_status
        self.valid_frequency_search = valid_frequency_search

        self.forced_fmax_lower_bound = forced_fmax_lower_bound
        self.forced_fmax_upper_bound = forced_fmax_upper_bound
        self.forced_custom_freq_list = forced_custom_freq_list

        self.overwrite = overwrite
        self.continue_on_error = continue_on_error
        self.force_single_thread = force_single_thread

        self.reset_lists()

        self.odatix_path = os.path.realpath(os.path.join(self.script_path, ".."))

    def reset_lists(self):
        self.checked_arch_param = []
        self.banned_arch_param = []
        self.valid_archs = []
        self.cached_archs = []
        self.overwrite_archs = []
        self.error_archs = []
        self.incomplete_archs = []
        self.new_archs = []
        self.deprecation_notice_archs = []

    def get_architectures(self, architectures, targets, constraint_filename="", install_path="", range_mode=False, keep=False, timestamp=""):

        self.reset_lists()
        self.architecture_instances = []

        only_one_target = len(targets) == 1

        # Define user accessible variables
        variables = Variables(
            tool_install_path=os.path.realpath(install_path),
            odatix_path=OdatixSettings.odatix_path,
            odatix_eda_tools_path=OdatixSettings.odatix_eda_tools_path,
        )
            
        with open(self.eda_target_filename, 'r') as f:
            try:
                settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
            except Exception as e:
                printc.error("Settings file \"" + self.eda_target_filename + "\" is not a valid YAML file", script_name)
                printc.cyan("error details: ", end="", script_name=script_name)
                print(str(e))
                sys.exit(-1)
            
            try:
                script_copy_enable = read_from_list('script_copy_enable', settings_data, self.eda_target_filename, type=bool, optional=True, script_name=script_name)
                if script_copy_enable:
                    script_copy_source = read_from_list('script_copy_source', settings_data, self.eda_target_filename, optional=True, script_name=script_name)        
                    script_copy_source = replace_variables(script_copy_source, variables) # Replace variables in command

                    if not os.path.isfile(script_copy_source):
                        printc.note("The script source file \"" + script_copy_source + "\" specified in \"" + self.eda_target_filename + "\" does not exist. Script copy disabled.", script_name)
                        raise BadValueInListError
                else:
                    raise BadValueInListError
            except (KeyNotInListError, BadValueInListError):
                script_copy_enable = False
                script_copy_source = "/dev/null"

            try:
                target_settings = read_from_list("target_settings", settings_data, self.eda_target_filename, optional=True, print_error=False, script_name=script_name)
            except (KeyNotInListError, BadValueInListError):
                target_settings = {}

            full_architectures = architectures
            for target in targets:
                # Overwrite existing script copy settings if there are target specific settings
                if target_settings != {}:
                    try:
                        this_target_settings = read_from_list(target, target_settings, self.eda_target_filename, optional=True, parent="target_settings", script_name=script_name)
                    except (KeyNotInListError, BadValueInListError):
                        this_target_settings = {}
                        pass
                    if this_target_settings != {}:
                        try:
                            script_copy_enable = read_from_list('script_copy_enable', this_target_settings, self.eda_target_filename, type=bool, optional=True, parent="target_settings/" + target, script_name=script_name)
                            if script_copy_enable:
                                script_copy_source = read_from_list('script_copy_source', this_target_settings, self.eda_target_filename, optional=True, parent="target_settings/" + target, script_name=script_name)        
                                script_copy_source = replace_variables(script_copy_source, variables) # Replace variables in command

                                if not os.path.isfile(script_copy_source):
                                    printc.note("The script source file \"" + script_copy_source + "\" specified in \"" + self.eda_target_filename + "\" does not exist. Script copy disabled.", script_name)
                                    raise BadValueInListError
                        except (KeyNotInListError, BadValueInListError):
                            script_copy_enable = False
                            script_copy_source = "/dev/null"

                # Handle wildcard
                architectures = ArchitectureHandler.configuration_wildcard(full_architectures, self.arch_path, target)

                for arch in architectures:
                    architecture_instance = self.get_architecture(
                        arch = arch,
                        target = target, 
                        only_one_target = only_one_target, 
                        script_copy_enable = script_copy_enable, 
                        script_copy_source = script_copy_source,
                        synthesis = True,
                        constraint_filename = constraint_filename,
                        install_path = install_path,
                        range_mode = range_mode,
                        keep=keep,
                        timestamp=timestamp,
                    )
                    if not range_mode:
                        if architecture_instance is not None:
                            self.architecture_instances.append(architecture_instance)
                    else:
                        if architecture_instance is not None:
                            for freq in architecture_instance.range_list:
                                freq_arch = copy.copy(architecture_instance)
                                formatted_freq = " {}@ {} MHz{}".format(printc.colors.GREY, freq, printc.colors.ENDC)
                                unformatted_display_name = freq_arch.arch_display_name
                                freq_arch.arch_display_name = freq_arch.arch_display_name + " @ " + str(freq) + " MHz"
                                freq_arch.tmp_dir = os.path.join(freq_arch.tmp_dir, str(freq) + "MHz")
                                freq_arch.tmp_script_path = os.path.join(freq_arch.tmp_dir, self.work_script_path)
                                freq_arch.tmp_report_path = os.path.join(freq_arch.tmp_dir, self.work_report_path)
                                freq_arch.tmp_log_path = os.path.join(freq_arch.tmp_dir, self.work_log_path)
                                freq_arch.target_frequency = freq
                                freq_arch.lib_name = freq_arch.lib_name + "_" + str(freq) + "MHz"

                                # check if the architecture is in cache and has a status file
                                status_file = os.path.join(freq_arch.tmp_dir, self.work_log_path, self.fmax_status_filename)
                                if isdir(freq_arch.tmp_dir) and isfile(status_file):
                                    # check if the previous synthesis has completed
                                    sf = open(status_file, "r")
                                    if self.valid_status in sf.read():
                                        if self.overwrite:
                                            printc.warning("Found cached results for \"" + unformatted_display_name + "\" @ " + str(freq) + " MHz with target \"" + target + "\".", script_name)
                                            self.overwrite_archs.append(unformatted_display_name)
                                            self.architecture_instances.append(freq_arch)
                                            self.valid_archs.append(unformatted_display_name + formatted_freq)
                                        else:
                                            printc.note("Found cached results for \"" + unformatted_display_name + "\" @ " + str(freq) + " MHz with target \"" + target + "\". Skipping.", script_name)
                                            self.cached_archs.append(freq_arch.arch_display_name)
                                    else: 
                                        printc.warning("The previous synthesis for \"" + unformatted_display_name + "\" @ " + str(freq) + " MHz with target \"" + target + "\" has not finished or the directory has been corrupted.", script_name)
                                        self.incomplete_archs.append(freq_arch.arch_display_name)
                                        self.architecture_instances.append(freq_arch)
                                        self.valid_archs.append(unformatted_display_name + formatted_freq)
                                    sf.close()
                                else:
                                    self.new_archs.append(unformatted_display_name + formatted_freq)
                                    self.architecture_instances.append(freq_arch)
                                    self.valid_archs.append(unformatted_display_name + formatted_freq)

        return self.architecture_instances

    @staticmethod
    def get_basic(arch, target="", only_one_target=True):
        arch_full = arch.replace(" ", "")

        parts = [part.strip() for part in arch_full.split('+')]

        arch = parts[0]
        requested_param_domains = parts[1:]

        if arch.endswith(".txt"):
            arch = arch[:-4] 
            printc.note("'.txt' after the configuration name is not needed. Just use \"" + arch + "\"", script_name)

        if arch.endswith("/"):
            arch = arch[:-1] 

        # get param dir (arch name before '/')
        arch_param_dir = re.sub('/.*', '', arch)

        if len(requested_param_domains) > 0: 
            arch_param_dir_work = arch_param_dir
            arch_display_name = arch + " [" + ", ".join(map(str, requested_param_domains)).replace("/", ":") + "]"
        else:
            arch_param_dir_work = arch_param_dir
            arch_display_name = arch_full

        if not only_one_target:
            arch_display_name = arch_display_name + " (" + target + ")"

        # get configuration (arch name after '/')
        arch_config = re.sub('.*/', '', arch)
        if len(requested_param_domains) > 0: 
            arch_config_dir_work = arch_config + "+" + "+".join(map(str, requested_param_domains)).replace("/", "_")
        else:
            arch_config_dir_work = arch_config

        return arch, arch_param_dir, arch_config, arch_display_name, arch_param_dir_work, arch_config_dir_work, requested_param_domains
    
    def get_architecture(self, arch, target="", only_one_target=True, script_copy_enable=False, script_copy_source="/dev/null", synthesis=False, constraint_filename="", install_path="", range_mode=False, keep=False, timestamp=""):
        
        arch, arch_param_dir, arch_config, arch_display_name, arch_param_dir_work, arch_config_dir_work, requested_param_domains = ArchitectureHandler.get_basic(arch, target, only_one_target)

        # check if there is a configuration specified
        if arch_config == arch_param_dir:
            # printc.note("No architecture configuration selected for \"" + arch +  "\". Using default parameters", script_name)
            arch = arch + "/" + arch
            no_configuration = True
        else:
            no_configuration = False
        
        arch_config_dir_work = arch_config_dir_work + "_" + timestamp if keep and timestamp != "" else arch_config_dir_work

        tmp_dir = os.path.join(self.work_path, target, arch_param_dir_work, arch_config_dir_work)
        fmax_status_file = os.path.join(tmp_dir, self.log_path, self.fmax_status_filename)
        frequency_search_file = os.path.join(tmp_dir, self.log_path, self.frequency_search_filename)

        # check if arch_param has been banned
        if arch_param_dir in self.banned_arch_param:
            self.error_archs.append(arch_display_name)
            return None

        # check if parameter dir exists
        arch_param = os.path.join(self.arch_path, arch_param_dir)
        if not isdir(arch_param):
            printc.error("There is no directory \"" + arch_param_dir + "\" in directory \"" + self.arch_path + "\"", script_name)
            self.banned_arch_param.append(arch_param_dir)
            self.error_archs.append(arch_display_name)
            return None
        
        # check if settings file exists
        if not isfile(os.path.join(arch_param, self.param_settings_filename)):
            printc.error("There is no setting file \"" + self.param_settings_filename + "\" in directory \"" + arch_param + "\"", script_name)
            self.banned_arch_param.append(arch_param_dir)
            self.error_archs.append(arch_display_name)
            return None

        # get settings variables
        settings_filename = os.path.join(self.arch_path, arch_param_dir, self.param_settings_filename)
        with open(settings_filename, 'r') as f:
            try:
                settings_data = yaml.load(f, Loader=yaml.loader.SafeLoader)
            except Exception as e:
                printc.error("Settings file \"" + settings_filename + "\" is not a valid YAML file", script_name)
                printc.cyan("error details: ", end="", script_name=script_name)
                print(str(e))
                self.banned_arch_param.append(arch_param_dir)
                self.error_archs.append(arch_display_name)
                return None # if an identifier is missing
            try:
                top_level_filename = read_from_list('top_level_file', settings_data, settings_filename, script_name=script_name)
                top_level_module   = read_from_list('top_level_module', settings_data, settings_filename, script_name=script_name)
                clock_signal       = read_from_list('clock_signal', settings_data, settings_filename, script_name=script_name)
                reset_signal       = read_from_list('reset_signal', settings_data, settings_filename, script_name=script_name)
            except (KeyNotInListError, BadValueInListError):
                self.banned_arch_param.append(arch_param_dir)
                self.error_archs.append(arch_display_name)
                return None # if an identifier is missing

            file_copy_enable, defined = get_from_dict('file_copy_enable', settings_data, settings_filename, type=bool, silent=True, default_value=False, script_name=script_name)
            if defined:
                file_copy_source, source_defined = get_from_dict('file_copy_source', settings_data, settings_filename, behavior=Key.MANTADORY, script_name=script_name)
                file_copy_dest, dest_defined = get_from_dict('file_copy_dest', settings_data, settings_filename, behavior=Key.MANTADORY, script_name=script_name)
                if not source_defined or not dest_defined:
                    self.banned_arch_param.append(arch_param_dir)
                    self.error_archs.append(arch_display_name)
                    return None
            else:
                file_copy_source = ""
                file_copy_dest = ""
            
            generate_command = ""
            generate_rtl, defined = get_from_dict('generate_rtl', settings_data, settings_filename, type=bool, silent=True, script_name=script_name)
            if not defined:
                generate_rtl = False

            if generate_rtl:
                local_rtl_path, _ = get_from_dict('generate_output', settings_data, settings_filename, default_value=self.work_rtl_path, script_name=script_name)
                rtl_path = self.work_rtl_path 
                generate_command, defined = get_from_dict('generate_command', settings_data, settings_filename, silent=True, script_name=script_name)
                if defined:
                    generate_rtl = True
                else:
                    printc.error("Cannot find key \"generate_command\" in \"" + settings_filename + "\" while generate_rtl=true", script_name)
                    self.banned_arch_param.append(arch_param_dir)
                    self.error_archs.append(arch_display_name)
                    generate_rtl = False
                    return None
            else:
                local_rtl_path = self.work_rtl_path 
                rtl_path, defined = get_from_dict('rtl_path', settings_data, settings_filename, script_name=script_name)
                if not defined:
                    self.banned_arch_param.append(arch_param_dir)
                    self.error_archs.append(arch_display_name)
                    return None

            top_level = os.path.join(rtl_path, top_level_filename)
            work_top_level = os.path.join(local_rtl_path, top_level_filename)

            use_parameters, start_delimiter, stop_delimiter, param_target_filename = self.get_use_parameters(arch, arch_display_name, settings_data, settings_filename, work_top_level, no_configuration, arch_param_dir=arch_param_dir)
            if use_parameters is None or start_delimiter is None or stop_delimiter is None or param_target_filename is None:
                return None

            design_path, design_path_defined = get_from_dict('design_path', settings_data, settings_filename, silent=True, script_name=script_name)
            if not defined:
                design_path = None
                if generate_rtl:
                    printc.error("Cannot find key \"design_path\" in \"" + settings_filename + "\" while generate_rtl=true", script_name)
                    self.banned_arch_param.append(arch_param_dir)
                    return None
            
            design_path_whitelist, _ = get_from_dict("design_path_whitelist", settings_data, settings_filename, type=list, default_value=[], silent=True, script_name=script_name)
            design_path_blacklist, _ = get_from_dict("design_path_blacklist", settings_data, settings_filename, type=list, default_value=[], silent=True, script_name=script_name)
            # if param_target_file is None:
            #   printc.error("Cannot find key \"param_target_file\" in \"" + settings_filename + "\" while generate_rtl=true", script_name)
            #   printc.note("\"param_target_file\" is the file in which parameters will be replaced before running the generate command", script_name)

        if not generate_rtl:
            # check if rtl path exists
            if not isdir(rtl_path):
                printc.error("The rtl path \"" + rtl_path + "\" specified in \"" + settings_filename + "\" does not exist", script_name)
                self.banned_arch_param.append(arch_param_dir)
                self.error_archs.append(arch_display_name)
                return None

            # check if top level file path exists
            if not isfile(top_level):
                printc.error("The top level file \"" + top_level_filename + "\" specified in \"" + settings_filename + "\" does not exist", script_name)
                self.banned_arch_param.append(arch_param_dir)
                self.error_archs.append(arch_display_name)
                return None

            # check if the top level module name exists in the top level file, at least
            f = open(top_level, "r")
            if top_level_module not in f.read():
                printc.error("There is no occurence of top level module name \"" + top_level_module + "\" in top level file \"" + top_level + "\"", script_name)
                self.banned_arch_param.append(arch_param_dir)
                self.error_archs.append(arch_display_name)
                f.close()
                return None
            f.close()
            
            # check if the top clock name exists in the top level file, at least
            f = open(top_level, "r")
            if clock_signal not in f.read():
                printc.error("There is no occurence of clock signal name \"" + clock_signal + "\" in top level file \"" + top_level + "\"", script_name)
                self.banned_arch_param.append(arch_param_dir)
                self.error_archs.append(arch_display_name)
                f.close()
                return None
            f.close()
            
            # check if the top reset name exists in the top level file, at least
            f = open(top_level, "r")
            if clock_signal not in f.read():
                printc.error("There is no occurence of reset signal name \"" + reset_signal + "\" in top level file \"" + top_level + "\"", script_name)
                self.banned_arch_param.append(arch_param_dir)
                self.error_archs.append(arch_display_name)
                f.close()
                return None
            f.close()

        # check if param file exists
        if not no_configuration:
            if not isfile(os.path.join(self.arch_path, arch + '.txt')):
                printc.error("The parameter file \"" + arch + ".txt\" does not exist in directory \"" + os.path.join(self.arch_path, arch_param_dir) + "\"", script_name)
                self.banned_arch_param.append(arch_param_dir)
                self.error_archs.append(arch_display_name)
                return None
        
        if len(requested_param_domains) > 0:
            param_domains = ParamDomain.get_param_domains(
                requested_param_domains=requested_param_domains, 
                architecture=arch_param_dir, 
                arch_path=self.arch_path, 
                param_settings_filename=self.param_settings_filename, 
                top_level_file=work_top_level
            )
            if param_domains is None:
                self.banned_arch_param.append(arch_param_dir)
                self.error_archs.append(arch_display_name)
                return None
        else:
            param_domains = []
            
        # optional settings
        formatted_bound = ""
        fmax_lower_bound = 0
        fmax_upper_bound = 0
        range_list = []

        if synthesis:
            fmax_lower_bound, fmax_upper_bound, range_list, warn_fmax_obsolete = self.get_frequency_settings(
                arch_config=arch_config,
                target=target, 
                settings_data=settings_data, 
                settings_filename=settings_filename, 
                range_mode=range_mode
            )

            # Override by bounds from --from and --to if used
            if self.forced_fmax_lower_bound is not None:
                fmax_lower_bound = self.forced_fmax_lower_bound
            if self.forced_fmax_upper_bound is not None:
                fmax_upper_bound = self.forced_fmax_upper_bound
            if self.forced_custom_freq_list is not None and self.forced_custom_freq_list != []:
                range_list = self.forced_custom_freq_list

            if warn_fmax_obsolete and not arch_param_dir in self.deprecation_notice_archs:
                self.deprecation_notice_archs.append(arch_param_dir)
                printc.warning("{} -> 'fmax_lower_bound' and 'fmax_upper_bound' are deprecated".format(settings_filename), script_name)
                printc.note("Use this syntax instead:", script_name)
                printc.magenta("fmax_synthesis:")
                printc.magenta("  lower_bound: XXX")
                printc.magenta("  upper_bound: XXX")

            # check if frequencies are valid
            if range_mode:
                if range_list is None or range_list == []: 
                    self.banned_arch_param.append(arch_param_dir)
                    self.error_archs.append(arch_display_name)
                    return None
            else:
                if not ArchitectureHandler.check_bounds(fmax_lower_bound, fmax_upper_bound):
                    self.banned_arch_param.append(arch_param_dir)
                    self.error_archs.append(arch_display_name)
                    return None

            fmax_lower_bound = str(fmax_lower_bound)
            fmax_upper_bound = str(fmax_upper_bound)

            formatted_bound = " {}({} - {} MHz){}".format(printc.colors.GREY, fmax_lower_bound, fmax_upper_bound, printc.colors.ENDC)

            # check if the architecture is in cache and has a status file
            if not range_mode:
                if isdir(tmp_dir) and isfile(fmax_status_file) and isfile(frequency_search_file):
                    # check if the previous synth_fmax has completed
                    sf = open(fmax_status_file, "r")
                    if self.valid_status in sf.read():
                        ff = open(frequency_search_file, "r")
                        if self.valid_frequency_search in ff.read():
                            if self.overwrite:
                                printc.warning("Found cached results for \"" + arch + "\" with target \"" + target + "\".", script_name)
                                self.overwrite_archs.append(arch_display_name + formatted_bound)
                            else:
                                printc.note("Found cached results for \"" + arch + "\" with target \"" + target + "\". Skipping.", script_name)
                                self.cached_archs.append(arch_display_name)
                                return None
                        else:
                            printc.warning("The previous synthesis for \"" + arch + "\" did not result in a valid maximum operating frequency.", script_name)
                            self.incomplete_archs.append(arch_display_name + formatted_bound)
                        ff.close()
                    else: 
                        printc.warning("The previous synthesis for \"" + arch + "\" has not finished or the directory has been corrupted.", script_name)
                        self.incomplete_archs.append(arch_display_name + formatted_bound)
                    sf.close()
                else:
                    if not range_mode:
                        self.new_archs.append(arch_display_name + formatted_bound)


        # Retrieve target-specific settings if they exist
        target_specific_data, target_specific_defined = get_from_dict(target, settings_data, settings_filename, silent=True)

        # target specific file copy
        if target_specific_defined:
            try:
                _file_copy_enable = read_from_list('file_copy_enable', target_specific_data, settings_filename, optional=True, print_error=False, type=bool, script_name=script_name)
                try:
                    _file_copy_source = read_from_list('file_copy_source', target_specific_data, settings_filename, optional=True, script_name=script_name)
                    _file_copy_dest = read_from_list('file_copy_dest', target_specific_data, settings_filename, optional=True, script_name=script_name)
                    file_copy_enable = _file_copy_enable
                    file_copy_source = _file_copy_source
                    file_copy_dest = _file_copy_dest
                except (KeyNotInListError, BadValueInListError):
                    pass
            except KeyNotInListError:
                pass
            except BadValueInListError:
                printc.note("Value \"" + str(_file_copy_enable) + "\" for key \"" + 'file_copy_enable' + "\"" + ", inside list \"" + "target_settings/" + target + "\"," + " in \"" + settings_filename + "\" is of type \"" + _file_copy_enable.__class__.__name__ + "\" while it should be of type \"bool\". Using default values instead.", script_name)
        
        # Define user accessible variables
        variables = Variables(
            tool_install_path=os.path.realpath(install_path),
            odatix_path=OdatixSettings.odatix_path,
            odatix_eda_tools_path=OdatixSettings.odatix_eda_tools_path,
        )

        # Replace variables in command
        file_copy_source = replace_variables(file_copy_source, variables)

        # check file copy
        if file_copy_enable:
            if not isfile(file_copy_source):
                printc.error("The source file to copy \"" + file_copy_source + "\" does not exist", script_name)
                self.banned_arch_param.append(arch_param_dir)
                self.error_archs.append(arch_display_name)
                return None

        # passed all check: added to the list
        if not range_mode:
            self.valid_archs.append(arch_display_name)
            self.checked_arch_param.append(arch_param_dir)

        lib_name = "LIB_" + target + "_" + arch_param_dir_work + "_" + arch_config_dir_work

        tmp_script_path = os.path.join(tmp_dir, self.work_script_path)
        tmp_report_path = os.path.join(tmp_dir, self.work_report_path)
        tmp_log_path = os.path.join(tmp_dir, self.work_log_path)

        arch_instance = Architecture(
            arch_name=arch,
            arch_display_name=arch_display_name,
            lib_name=lib_name,
            target=target,
            local_rtl_path=local_rtl_path,
            tmp_script_path=tmp_script_path,
            tmp_log_path=tmp_log_path,
            tmp_report_path=tmp_report_path,
            tmp_dir=tmp_dir,
            design_path=design_path,
            design_path_whitelist=design_path_whitelist,
            design_path_blacklist=design_path_blacklist,
            rtl_path=rtl_path,
            log_path=tmp_log_path,
            arch_path=self.arch_path,
            clock_signal=clock_signal,
            reset_signal=reset_signal,
            top_level_module=top_level_module,
            top_level_filename=top_level_filename,
            file_copy_enable=file_copy_enable,
            file_copy_source=file_copy_source,
            file_copy_dest=file_copy_dest,
            script_copy_enable = script_copy_enable,
            script_copy_source = script_copy_source,
            fmax_lower_bound=fmax_lower_bound,
            fmax_upper_bound=fmax_upper_bound,
            range_list=range_list,
            target_frequency=0,
            param_target_filename=param_target_filename,
            generate_rtl=generate_rtl,
            use_parameters=use_parameters,
            start_delimiter=start_delimiter,
            stop_delimiter=stop_delimiter,
            generate_command=generate_command,
            constraint_filename=constraint_filename,
            install_path=install_path,
            param_domains=param_domains,
            continue_on_error=self.continue_on_error,
            force_single_thread=self.force_single_thread,
        )

        return arch_instance

    def get_use_parameters(self, arch, arch_display_name, settings_data, settings_filename, top_level_file, no_configuration=False, add_to_error_list=True, arch_param_dir=""):

        use_parameters, start_delimiter, stop_delimiter, param_target_filename = ParamDomain.get_param_delimiters(settings_data, settings_filename, top_level_file)

        if no_configuration:
            use_parameters = False
        else:
            if use_parameters:
                # check if parameter file exists
                param_file = os.path.join(self.arch_path, arch + ".txt")
                if not isfile(param_file):
                    printc.error("There is no parameter file \"" + param_file + "\", while \"use_parameters\" is true", script_name)
                    if add_to_error_list:
                        self.error_archs.append(arch_display_name)
                    return None, None, None, None

        return use_parameters, start_delimiter, stop_delimiter, param_target_filename

    @staticmethod
    def get_frequency_settings(arch_config, target, settings_data, settings_filename, range_mode):
        """
        Retrieves frequency synthesis settings from the YAML configuration.

        Args:
                arch_config (str): The architecture configuration.
                target (str): The target FPGA/ASIC.
                settings_data (dict): The parsed YAML settings data.
                settings_filename (str): Name of the YAML file.
                range_mode (bool): Whether to use custom frequency synthesis.

        Returns:
                tuple: (fmax_lower_bound, fmax_upper_bound, custom_freq_list, warn_fmax_obsolete).
        """

        # Defaults
        fmax_lower_bound = None
        fmax_upper_bound = None
        custom_freq_list = []
        warn_fmax_obsolete = False

        target_fmax_defined = False
        target_custom_freq_defined = False
        arch_specific_defined = False
        arch_fmax_defined = False
        arch_custom_freq_defined = False

        # Retrieve general settings
        global_fmax_data, global_fmax_defined =  get_from_dict("fmax_synthesis", settings_data, settings_filename, silent=True)
        global_custom_freq_data, global_custom_freq_defined = get_from_dict("custom_freq_synthesis", settings_data, settings_filename, silent=True)

        # Retrieve target-specific settings
        target_specific_data, target_specific_defined = get_from_dict(target, settings_data, settings_filename, silent=True)
        if target_specific_defined:
            target_fmax_data, target_fmax_defined =  get_from_dict("fmax_synthesis", target_specific_data, settings_filename, silent=True)
            target_custom_freq_data, target_custom_freq_defined = get_from_dict("custom_freq_synthesis", target_specific_data, settings_filename, silent=True)

            # Retrieve architecture-specific settings
            arch_specific_data, arch_specific_defined = get_from_dict(arch_config, target_specific_data, settings_filename, silent=True)
            if arch_specific_defined:
                arch_fmax_data, arch_fmax_defined =  get_from_dict("fmax_synthesis", arch_specific_data, settings_filename, silent=True)
                arch_custom_freq_data, arch_custom_freq_defined = get_from_dict("custom_freq_synthesis", arch_specific_data, settings_filename, silent=True)

        if range_mode: # Custom frequency synthesis

            # Get lower bound
            lower_bound_defined = False
            if arch_custom_freq_defined:
                range_lower_bound, lower_bound_defined = get_from_dict("lower_bound", arch_custom_freq_data, settings_filename, default_value=None, silent=True)
            if not lower_bound_defined and target_custom_freq_defined:
                range_lower_bound, lower_bound_defined = get_from_dict("lower_bound", target_custom_freq_data, settings_filename, default_value=None, silent=True)
            if not lower_bound_defined and global_custom_freq_defined:
                range_lower_bound, lower_bound_defined = get_from_dict("lower_bound", global_custom_freq_data, settings_filename, default_value=None, silent=True)

            # Get upper bound
            upper_bound_defined = False
            if arch_custom_freq_defined:
                range_upper_bound, upper_bound_defined = get_from_dict("upper_bound", arch_custom_freq_data, settings_filename, default_value=None, silent=True)
            if not upper_bound_defined and target_custom_freq_defined:
                range_upper_bound, upper_bound_defined = get_from_dict("upper_bound", target_custom_freq_data, settings_filename, default_value=None, silent=True)
            if not upper_bound_defined and global_custom_freq_defined:
                range_upper_bound, upper_bound_defined = get_from_dict("upper_bound", global_custom_freq_data, settings_filename, default_value=None, silent=True)

            # Get step
            step_defined = False
            if arch_custom_freq_defined:
                range_step, step_defined = get_from_dict("step", arch_custom_freq_data, settings_filename, default_value=None, silent=True)
            if not step_defined and target_custom_freq_defined:
                range_step, step_defined = get_from_dict("step", target_custom_freq_data, settings_filename, default_value=None, silent=True)
            if not step_defined and global_custom_freq_defined:
                range_step, step_defined = get_from_dict("step", global_custom_freq_data, settings_filename, default_value=None, silent=True)

            # Get list
            list_defined = False
            list_defined_arch = False
            list_defined_target = False
            list_defined_global = False
            list_append = False
            if arch_custom_freq_defined:
                custom_freq_list, list_defined_arch = get_from_dict("list", arch_custom_freq_data, settings_filename, default_value=None, silent=True)
                if list_defined_arch:
                    list_append, _ = get_from_dict("list_append", arch_custom_freq_data, settings_filename, default_value=False, silent=True)
            if target_custom_freq_defined:
                tmp_list, list_defined_target = get_from_dict("list", target_custom_freq_data, settings_filename, default_value=None, silent=True)
                if list_defined_target:
                    if not list_defined_arch:
                        custom_freq_list = tmp_list
                    elif list_append:
                        custom_freq_list = custom_freq_list + tmp_list
                    list_append, _ = get_from_dict("list_append", target_custom_freq_data, settings_filename, default_value=False, silent=True)
            if global_custom_freq_defined:
                tmp_list, list_defined_global = get_from_dict("list", global_custom_freq_data, settings_filename, default_value=None, silent=True)
                if list_defined_global:
                    if not list_defined_target:
                        custom_freq_list = tmp_list
                    elif list_append:
                        custom_freq_list = custom_freq_list + tmp_list

            list_defined = list_defined_global or list_defined_target or list_defined_arch
            if not list_defined:
                custom_freq_list = []

            if not step_defined or (step_defined and (range_step == 0 or range_step == False)): # Check if range is deactivated
                range_list = []
            else:
                if lower_bound_defined and upper_bound_defined and step_defined:
                    if ArchitectureHandler.check_bounds(range_lower_bound, range_upper_bound, range_step, synth_type="custom frequency synthesis"):
                        range_list = ArchitectureHandler.create_list_from_range(range_lower_bound, range_upper_bound, range_step)
                        custom_freq_list = custom_freq_list + range_list

            # Check if a list is defined
            if len(custom_freq_list) == 0:
                # printc.error('Could not find any valid custom frequency definition in "{}" for architecture configuration "{}" with target "{}"'.format(settings_filename, arch_config, target), script_name)
                # printc.note('You can define custom synthesis frequencies like this:', script_name)
                # printc.magenta("custom_freq_synthesis:")
                # printc.magenta("  lower_bound: XXX")
                # printc.magenta("  upper_bound: XXX")
                # printc.magenta("  step: XXX")
                # printc.cyan("or ")
                # printc.magenta("custom_freq_synthesis:")
                # printc.magenta("  step: No")
                # printc.magenta("  list: [XXX, XXX, XXX]")
                # printc.cyan("or ")
                # printc.magenta("custom_freq_synthesis:")
                # printc.magenta("  lower_bound: XXX")
                # printc.magenta("  upper_bound: XXX")
                # printc.magenta("  step: XXX")
                # printc.magenta("  list: [XXX, XXX, XXX] # append to the list generated by range")
                custom_freq_list = hard_settings.default_custom_freq_list

            return None, None, custom_freq_list, warn_fmax_obsolete

        else: # fmax synthesis

            # Legacy fallback for older odatix version
            lower_defined = False
            if target_specific_defined:
                fmax_lower_bound, lower_defined = get_from_dict("fmax_lower_bound", target_specific_data, settings_filename, silent=True)
            if not lower_defined and arch_specific_defined:
                fmax_lower_bound, lower_defined = get_from_dict("fmax_lower_bound", arch_specific_data, settings_filename, silent=True)
            upper_defined = False
            if target_specific_defined:
                fmax_upper_bound, upper_defined = get_from_dict("fmax_upper_bound", target_specific_data, settings_filename, silent=True)
            if not upper_defined and arch_specific_defined:
                fmax_upper_bound, upper_defined = get_from_dict("fmax_upper_bound", arch_specific_data, settings_filename, silent=True)
            
            # Deprecation warning
            if lower_defined or upper_defined:
                warn_fmax_obsolete = True

            # Get lower bound
            defined = False
            if arch_fmax_defined:
                fmax_lower_bound, defined = get_from_dict("lower_bound", arch_fmax_data, settings_filename, default_value=None, silent=True)
            if not defined and target_fmax_defined:
                fmax_lower_bound, defined = get_from_dict("lower_bound", target_fmax_data, settings_filename, default_value=None, silent=True)
            if not defined and global_fmax_defined:
                fmax_lower_bound, defined = get_from_dict("lower_bound", global_fmax_data, settings_filename, default_value=None, silent=True)

            # Get upper bound
            defined = False
            if arch_fmax_defined:
                fmax_upper_bound, defined = get_from_dict("upper_bound", arch_fmax_data, settings_filename, default_value=None, silent=True)
            if not defined and target_fmax_defined:
                fmax_upper_bound, defined = get_from_dict("upper_bound", target_fmax_data, settings_filename, default_value=None, silent=True)
            if not defined and global_fmax_defined:
                fmax_upper_bound, defined = get_from_dict("upper_bound", global_fmax_data, settings_filename, default_value=None, silent=True)

            # Check if bounds are defined
            if fmax_lower_bound is None:
                fmax_lower_bound = hard_settings.default_fmax_lower_bound
                # printc.error('Lower bound for fmax synthesis is not defined in "{}" for architecture configuration "{}" with target "{}"'.format(settings_filename, arch_config, target), script_name)
            if fmax_upper_bound is None:
                fmax_upper_bound = hard_settings.default_fmax_upper_bound
                # printc.error('Upper bound for fmax synthesis is not defined in "{}" for architecture configuration "{}" with target "{}"'.format(settings_filename, arch_config, target), script_name)
            # if fmax_lower_bound is None or fmax_upper_bound is None:
            #     printc.note('You can define fmax synthesis frequency bounds like this:', script_name)
            #     printc.magenta("fmax_synthesis:")
            #     printc.magenta("  lower_bound: XXX")
            #     printc.magenta("  upper_bound: XXX")
            #     return None, None, None, warn_fmax_obsolete

            return fmax_lower_bound, fmax_upper_bound, None, warn_fmax_obsolete

    @staticmethod
    def configuration_wildcard(full_architectures, arch_path=OdatixSettings.DEFAULT_ARCH_PATH, target=""):
        architectures = []
        joker_archs = []
        for arch in full_architectures:
            arch, arch_param_dir, arch_config, _, _, _, requested_param_domains = ArchitectureHandler.get_basic(arch, target, False)
            if arch.endswith("/*"):
                # get param dir (arch name before '/*')
                arch_param_dir = re.sub(r'/\*', '', arch)
                arch_param = os.path.join(arch_path, arch_param_dir)
                # check if parameter dir exists
                if os.path.isdir(arch_param):
                    files = [f[:-4] for f in os.listdir(arch_param) if os.path.isfile(os.path.join(arch_param, f)) and f.endswith(".txt")]
                    joker_archs = [os.path.join(arch_param_dir, file) for file in sorted(files)]
                    if len(joker_archs) == 0:
                        printc.note(f"No configuration found in \"{arch_param}\"", script_name)
                else:
                    printc.error(f"The architecture directory \"{arch_param}\" does not exist", script_name)
            else:
                joker_archs = [arch]
                arch_param = arch_param_dir
                
            # Parameter domain wildcard
            joker_param_domain = {}
            if len(requested_param_domains) > 0:
                for requested_param_domain in requested_param_domains:
                    if requested_param_domain.endswith("/*"):
                        param_domain = re.sub(r'/\*', '', requested_param_domain)
                        # get parameter domain dir
                        arch_param = os.path.join(arch_path, arch_param_dir)
                        if not os.path.isdir(arch_param):
                            printc.error(f"The architecture directory \"{arch_param}\" does not exist", script_name)
                            continue
                        param_domain_dir = os.path.join(arch_param, param_domain)  
                        # check if parameter domain dir exists
                        if os.path.isdir(param_domain_dir):
                            files = [f[:-4] for f in os.listdir(param_domain_dir) if os.path.isfile(os.path.join(param_domain_dir, f)) and f.endswith(".txt")]
                            joker_param_domain[param_domain] = sorted(files)
                        else:
                            printc.error(f"The parameter domain directory \"{param_domain_dir}\" does not exist", script_name)
                            existing_domains = [d for d in os.listdir(arch_param) if os.path.isdir(os.path.join(arch_param, d))]
                            if len(existing_domains) == 0:
                                printc.tip(f"No parameter domains found in \"{arch_param}\"", script_name)
                            else:
                                printc.tip(f"Available parameter domains found in \"{arch_param}\": {', '.join(existing_domains)}", script_name)
                            continue
                    else:
                        param_domain = re.sub(r'/.*', '', requested_param_domain)
                        value = re.sub(r'.*/', '', requested_param_domain)
                        joker_param_domain[param_domain] = [value]  
                if len(joker_param_domain) == 0:
                    return None  
                # Generate combinations
                param_keys = list(joker_param_domain.keys())
                param_values = [joker_param_domain[key] if isinstance(joker_param_domain[key], list) else [joker_param_domain[key]] for key in param_keys]  
                for arch_instance in joker_archs:
                    for param_combination in itertools.product(*param_values):
                        param_string = "+".join(f"{param_keys[i]}/{param_combination[i]}" for i in range(len(param_keys)))
                        architectures.append(f"{arch_instance}+{param_string}")
            else:
                architectures = architectures + joker_archs  
        # Remove duplicates
        architectures = list(dict.fromkeys(architectures))
        return architectures

    def create_list_from_range(lower_bound, upper_bound, step):
        return list(range(lower_bound, upper_bound + 1, step))

    def check_bounds(lower_bound, upper_bound, step=0, synth_type="fmax synthesis"):
        success = True
        if not isinstance(lower_bound, int):
            printc.error('Lower bound for {} is "{}" which is a "{}" while it should be an integer'.format(synth_type, lower_bound, type(lower_bound).__name__), script_name)
            success =  False
        if not isinstance(upper_bound, int):
            printc.error('Upper bound for {} is "{}" which is a "{}" while it should be an integer'.format(synth_type, upper_bound, type(upper_bound).__name__), script_name)
            success =  False
        if not isinstance(step, int):
            printc.error('Step for {} is "{}" which is a "{}" while it should be an integer or "No"'.format(synth_type, upper_bound, type(upper_bound).__name__), script_name)
            success =  False
        if success:
            if upper_bound <= lower_bound:
                printc.error("The upper bound ({}) for {} must be strictly greater than the lower bound ({})".format(synth_type, upper_bound, lower_bound), script_name)
                success =  False
        return success

    def print_summary(self):
        ArchitectureHandler.print_arch_list(self.cached_archs, "Existing results (skipped -> use '-o' to overwrite)", printc.colors.CYAN)
        ArchitectureHandler.print_arch_list(self.overwrite_archs, "Existing results (will be overwritten)", printc.colors.YELLOW)
        ArchitectureHandler.print_arch_list(self.incomplete_archs, "Incomplete results (will be overwritten)", printc.colors.YELLOW)
        ArchitectureHandler.print_arch_list(self.new_archs, "New architectures", printc.colors.ENDC)
        ArchitectureHandler.print_arch_list(self.error_archs, "Invalid settings, (skipped, see errors above)", printc.colors.RED)

    def get_valid_arch_count(self):
        return len(self.valid_archs)

    @staticmethod
    def print_arch_list(arch_list, description, color):
        if not len(arch_list) > 0:
            return

        print()
        printc.bold(description + ":")
        for arch in arch_list:
            printc.color(color)
            print("  - " + arch)
        printc.endc()
